/*
 * out_yasapi-nt.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <windows.h>
#include <shlobj.h>
#include <yasapi.h>
#include <winamp/wa_ipc.h>
#define YASAPI_EXPORT
#include <yasapi_nt_api.h>

static int SubclassWinamp(HWND hWnd);
static int UnSubclassWinamp(HWND hWnd);

int getwrittentime();
int getoutputtime();

static Player player;
static Out_Module out;

BOOL WINAPI _DllMainCRTStartup(HANDLE hInst, ULONG ul_reason_for_call, LPVOID lpReserved)
{
  return TRUE;
}

void config(HWND hwnd)
{
  GtkClientShow(&player.gtk,&player.gtk.config,hwnd);
}

void about(HWND hwnd)
{
  GtkClientShow(&player.gtk,&player.gtk.about,hwnd);
}

void init()
{
  DMARKLN();

  if (PlayerCreate(&player,&out)<0) {
    DMESSAGE("creating player");
    goto player;
  }

  if (SubclassWinamp(out.hMainWindow)<0) {
    DMESSAGE("subclassing winamp");
    goto subclass;
  }

  DMARKLN();

  return;
//cleanup:
subclass:
  DMARKLN();
  PlayerDestroy(&player);
player:
  DMARKLN();
  return;
}

void quit()
{
  UnSubclassWinamp(out.hMainWindow);
  PlayerDestroy(&player);
}

int open(int samplerate, int numchannels, int bitspersamp, int bufferlenms, int prebufferms)
{
  int latency;

  DMARKLN();

  if (PlayerOpen(&player,samplerate,numchannels,bitspersamp,&latency,NULL)<0)
    return -1;

  DVWRITELN("latency: %d",latency);

  return latency;
}

void close()
{
DVWRITELN("%d",eEngineClientStateOpen<player.engine.fsm.pState-player.engine.fsm.pStateMin);
  EngineClientClose(&player.engine);
DMARKLN(); DMARKLN(); DMARKLN(); DMARKLN();
}

int write(char *buf, int len)
{
  // 0 on success. Len == bytes to write (<= 8192 always). buf is straight
  //   audio data. 
  // 1 returns not able to write (yet). Non-blocking, always.
  int code;

  code=EngineClientWrite(&player.engine,(BYTE *)buf,len);

  if (code<0)
    return 1;

  return 0;
}

int canwrite()
{
  int code;
  code=EngineClientAvailable(&player.engine);
  return code;
}

int isplaying()
{
int code;

DWRITELN("[");
  code=EngineClientIsPlaying(&player.engine);
DWRITELN("]");

  return code;
}

int pause(int pause)
{
  return EngineClientPause(&player.engine,pause);
}

void setvolume(int volume)
{
  PlayerVolume(&player,volume);
}

void setpan(int pan)
{
}

void flush(int t)
{
  EngineClientFlush(&player.engine,t);
}
  
int getoutputtime()
{
  return EngineClientOutputTime(&player.engine);
}

int getwrittentime()
{
  return EngineClientWrittenTime(&player.engine);
}

static Out_Module out = {
  OUT_VER,
  YASAPI_TITLE,
  65536+128,
  0, // hmainwindow
  0, // hdllinstance
  config,
  about,
  init,
  quit,
  open,
  close,
  write,
  canwrite,
  isplaying,
  pause,
  setvolume,
  setpan,
  flush,
  getoutputtime,
  getwrittentime
};

__declspec( dllexport ) Out_Module * winampGetOutModule()
{
  return &out;
}

///////////////////////////////////////////////////////////////////////////////
static WNDPROC WinampProc;

static LRESULT CALLBACK PluginWinampProc(HWND hWnd, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
  char *modulea;

  switch (uMsg) {
  case WM_CLOSE:
    YASAPI_FSM_DOWN(&player.present,0);
    break;
  case WM_WA_IPC:
    switch (lParam) {
    case IPC_CB_OUTPUTCHANGED:
      modulea=(char *)SendMessageW(hWnd,WM_WA_IPC,0,IPC_GETOUTPUTPLUGIN);
      player.engine.param.bActive=!yasapi_wcsstrcmp(player.pBasename,modulea);
      break;
    default:
      break;
    }

    break;
  default:
    break;
  }

  return CallWindowProc(
    WinampProc,   // _In_ WNDPROC lpPrevWndFunc,
    hWnd,         // _In_ HWND    hWnd,
    uMsg,         // _In_ UINT    Msg,
    wParam,       // _In_ WPARAM  wParam,
    lParam        // _In_ LPARAM  lParam
  );
}

static int SubclassWinamp(HWND hWnd)
{
  if (NULL==hWnd||NULL!=WinampProc) {
    DMESSAGE("failing to subclass winamp window");
    return -1;
  }

  WinampProc=(WNDPROC)SetWindowLong(
    hWnd,                   // _In_ HWND hWnd,
    GWL_WNDPROC,            // _In_ int  nIndex,
    (LONG)PluginWinampProc  // _In_ LONG dwNewLong
  );

  return 0;
}

static int UnSubclassWinamp(HWND hWnd)
{
  if (NULL==hWnd||NULL==WinampProc) {
    DMESSAGE("failing to unsubclass winamp window");
    return -1;
  }

  SetWindowLong(
    hWnd,                   // _In_ HWND hWnd,
    GWL_WNDPROC,            // _In_ int  nIndex,
    (LONG)WinampProc        // _In_ LONG dwNewLong
  );

  return 0;
}

///////////////////////////////////////////////////////////////////////////////
static EYasapiNtApi YasapiNtApiVersion(void)
{
  return eYasapiNtApiVersion;
}

static int YasapiNtApiOpen(int samplerate, int numchannels, int bitspersamp,
    int *pnMaxLatency, int *pnFramesDevice)
{
  int code;

DWRITELN("[");
  code=PlayerOpen(&player,samplerate,numchannels,bitspersamp,pnMaxLatency,
      pnFramesDevice);
DWRITELN("]");

  return code;
}

static int YasapiNtApiWrite(int msTimeout, const char *buf, int len)
{
  int code;

  code=EngineClientApiWrite(&player.engine,msTimeout,buf,len);

  return code;
}

static int YasapiNtApiEndOfTrack(int msTimeout)
{

  int code;

DWRITELN("[");
  code=EngineClientApiEndOfTrack(&player.engine,msTimeout);
DWRITELN("]");

  return code;
}

static void YasapiNtApiKill(void)
{
DWRITELN("[");
  EngineClientApiKill(&player.engine);
DWRITELN("]");
}

#define YASPI_EXPORT_API
#if defined (YASPI_EXPORT_API) // [
YASAPI_EXPORT
#endif // ]
const YasapiNtApi *GetYasapiNtApi(void)
{
  static const YasapiNtApi api={
    .Version=YasapiNtApiVersion,
    .v1.Open=YasapiNtApiOpen,
    .v1.Write=YasapiNtApiWrite,
    .v1.EndOfTrack=YasapiNtApiEndOfTrack,
    .v1.Kill=YasapiNtApiKill,
  };

  return &api;
}
