/*
 * yasapi.h
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#ifndef __YASAPI_H__ // [
#define __YASAPI_H__

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <math.h>
#include <yasapi_param.h>

#ifdef __cplusplus // [
extern "C" {
#endif // ]


///////////////////////////////////////////////////////////////////////////////
typedef int32_t sample_t;

typedef const struct _IRingDitherVtbl IRingDitherVtbl;
typedef enum _ERingFormat ERingFormat;
typedef struct _Ring Ring;
typedef enum _EEngineServer EEngineServer;
typedef const struct _IEngineServerState IEngineServerState;
typedef const struct _IEngineServerTimeVtbl IEngineServerTimeVtbl;
typedef struct _EngineServerTime EngineServerTime;
typedef const struct _IEngineServerVolumeVtbl IEngineServerVolumeVtbl;
typedef struct _EngineServerVolumeSimple EngineServerVolumeSimple;
typedef struct _EngineServerVolumeEndpoint EngineServerVolumeEndpoint;
typedef struct _EngineServerVolume EngineServerVolume;
typedef struct _EngineServer EngineServer;
typedef struct _EngineOptions EngineOptions;
typedef struct _EngineWfx EngineWfx;
typedef struct _EngineParam EngineParam;
typedef struct _EngineMonitor EngineMonitor;
typedef struct _EngineStats EngineStats;
typedef enum _EEngineClient EEngineClient;
typedef const struct _IEngineClientState IEngineClientState;
typedef struct _EngineClient EngineClient;
typedef int RingCreateFun(Ring *pRing, EngineClient *pClient, UINT32 uFrames);
typedef enum _EPresentServerState EPresentServerState;
typedef const struct _IPresentServerState IPresentServerState;
typedef struct _PresentServer PresentServer;
typedef enum _EPresentClientState EPresentClientState;
typedef const struct _IPresentClientState IPresentClientState;
typedef struct _PresentClient PresentClient;
typedef enum _EGtkServer EGtkServer;
typedef enum _EGtkLib EGtkLib;
typedef const struct _IGtkServerState IGtkServerState;
typedef enum _EGtkClient EGtkClient;
typedef enum _EPlayerState EPlayerState;
typedef const struct _IPlayerState IPlayerState;
typedef enum _EPlayerLimits EPlayerLimits;

///////////////////////////////////////////////////////////////////////////////
struct _IRingDitherVtbl {
  const char *pId;
  void (*Start)(Ring *pRing);
  sample_t (*Noise)(Ring *pRing, WORD nChannel);
};

extern IRingDitherVtbl gcIRingDitherNullVtbl;
extern IRingDitherVtbl gcIRingDitherRoundVtbl;
extern IRingDitherVtbl gcIRingDitherSimpleVtbl;

////////////////
enum _ERingFormat {
  eRingFormatBitsPerSample=32,
};


////////////////
struct _Ring {
  void (*Write)(Ring *pRing, const BYTE **ppSrc, DWORD dwSize);
  void (*Read)(Ring *pRing, BYTE **ppDst, DWORD dwSize);
  WAVEFORMATEXTENSIBLE wfx;

  struct {
    IRingDitherVtbl *lpVtbl;

    union {
      struct {
        sample_t noise;
      } round;

      struct {
        DWORD dwScaleNoise;
        sample_t noise;
      } simple;
    };
  } dither;

  EngineClient *pClient;
  BYTE *_buf;
  BYTE *_rp;
  BYTE *_wp;
  BYTE *_mp;

  struct {
    UINT32 _uWritten;
    UINT32 _uAvailable;
  } frames;
};

int RingCreateFrames(Ring *pRing, EngineClient *pClient, UINT32 uFrames);
void RingDestroy(Ring *pRing);

int RingReallocFrames(Ring *pRing, EngineClient *pClient, UINT32 uFrames);
void RingReset(Ring *pRing);
int RingResetWrite(Ring *pRing, int bInitial);
WORD RingFramesMaximum(Ring *pRing);
int RingWrite(Ring *pRing, const BYTE *buf, DWORD dwSizeSource);
int RingReadFrames(Ring *pRing, BYTE *buf, WORD uFrames);

DWORD RingSourceSize(Ring *pRing);
DWORD RingSourceSizeAvailable(Ring *pRing);
DWORD RingSourceSizeWritten(Ring *pRing);
DWORD RingTargetSize(Ring *pRing);
DWORD RingTargetSizeWritten(Ring *pRing);

////////////////
int WfxCreate(WAVEFORMATEXTENSIBLE *pWfx, DWORD samplerate, WORD numchannels,
    WORD bitspersamp, int bSurround);

///////////////////////////////////////////////////////////////////////////////
struct _IFsmState {
  const char *pId;
  int (*Enter)(void *pFsm);
  void (*Leave)(void *pFsm);
};

///////////////////////////////////////////////////////////////////////////////
enum _EEngineServer {
  eEngineServerStateStart,
#define YASAPI_ENGINE_SERVER(x) eEngineServerState ## x,
#include <yasapi_engine_server_state.h>
#undef YASAPI_ENGINE_SERVER
  eEngineServerStateMax,
  eEngineServerStateBase=eEngineServerStateApiEvent,
  eEngineServerStateOpen=eEngineServerStateRender,
  eEngineServerStateCount=eEngineServerStateMax-1,
};

struct _IEngineServerState {
  const char *pId;
  int (*Enter)(EngineServer *pServer);
  void (*Leave)(EngineServer *pServer);
};

extern IEngineServerState gcaIEngineServerState[eEngineServerStateCount];

///////////////////////////////////////////////////////////////////////////////
struct _IEngineServerTimeVtbl {
  const char *pId;
  int (*ClockCreate)(EngineServer *pServer);
  void (*ClockDestroy)(EngineServer *pServer);
  ////////////////
  void (*Inc)(EngineServer *pServer);
  int (*Get)(EngineServer *pServer);
  void (*Set)(EngineServer *pServer, int ms);
};

extern IEngineServerTimeVtbl gcIEngineServerTimeNullVtbl;
extern IEngineServerTimeVtbl gcIEngineServerTimeClockVtbl;

///////////////////////////////////////////////////////////////////////////////
struct _EngineServerTime {
  IEngineServerTimeVtbl *lpVtbl;
  double ms;
};

///////////////////////////////////////////////////////////////////////////////
struct _IEngineServerVolumeVtbl {
  const char *pId;
  int bApplyGain;
  int (*Create)(EngineServer *pServer);
  void (*Destroy)(EngineServer *pServer);
  void (*Set)(EngineServer *pServer);
};

extern IEngineServerVolumeVtbl gcIEngineServerVolumeNullVtbl;
extern IEngineServerVolumeVtbl gcIEngineServerVolumeSoftwareVtbl;
extern IEngineServerVolumeVtbl gcIEngineServerVolumeSimpleAudioVtbl;
extern IEngineServerVolumeVtbl gcIEngineServerVolumeAudioEndpointVtbl;

///////////////////////////////////////////////////////////////////////////////
struct _EngineServerVolumeSimple {
  ISimpleAudioVolume *pSimpleAudioVolume;
};

struct _EngineServerVolumeEndpoint {
  IAudioEndpointVolume *pAudioEndpointVolume;
  float fLevel;
  float fVolumeMindB;
  float fVolumeMaxdB;
  float fVolumeIncrementdB;
};

///////////////////////////////////////////////////////////////////////////////
struct _EngineServerVolume {
  IEngineServerVolumeVtbl *lpVtbl;

  union {
    EngineServerVolumeSimple simple;
    EngineServerVolumeEndpoint endpoint;
  };
};

///////////////////////////////////////////////////////////////////////////////
struct _EngineServer {
  Fsm fsm;
  EngineClient *pClient;

  struct {
    HANDLE hEvent;
    EEngineServer eStateRequest;
  } thread;


  struct _EngineClientApi {
    HANDLE hEvent;
  } api;

  struct {
    HANDLE hTask;
    DWORD dwTaskIndex;
  } task;

  IAudioClient *pAudioClient;
  IAudioClock *pAudioClock;
  IAudioRenderClient *pAudioRenderClient;

  struct {
    int bBoosted;
    int nPrevious;
    int bAltered;
  } priority;
};

DWORD WINAPI EngineServerThreadProc(LPVOID lParameter);

int EngineServerCreate(EngineServer *pServer, EngineClient *pClient);
void EngineServerDestroy(EngineServer *pServer);

int EngineServerRender(EngineServer *pServer, EEngineServer eEngineServer,
    EEngineServer eEngineServerRequest);
int _EngineServerRender(EngineServer *pServer, UINT32 uFrames, int bSilence);
int EngineServerNumFramesPadding(EngineServer *pServer,
    UINT32 *puNumFramesPadding);

///////////////////////////////////////////////////////////////////////////////
enum _EEngineClient {
  eEngineClientStateStart,
#define YASAPI_ENGINE_CLIENT(x) eEngineClientState ## x,
#include <yasapi_engine_client_state.h>
#undef YASAPI_ENGINE_CLIENT
  eEngineClientStateMax,
  eEngineClientStateOpen=eEngineClientStateRing,
  eEngineClientStateBase=eEngineClientStateThreadRunning,
  eEngineClientStatePreOpen=eEngineClientStateBufferSize,
  eEngineClientStateCount=eEngineClientStateMax-1,
};

///////////////////////////////////////////////////////////////////////////////
struct _EngineOptions {
  OptionsCommon common;
  OptionsDevice device;
};

int EngineOptionsCreate(EngineOptions *pOptions, EngineClient *pClient);

///////////////////////////////////////////////////////////////////////////////
struct _EngineWfx {
  WAVEFORMATEXTENSIBLE source;
  WAVEFORMATEXTENSIBLE target;
};

int EngineWfxCreate(EngineWfx *pWfx, const EngineOptions *pOptions,
    int samplerate, int numchannels, int bitspersamp);

///////////////////////////////////////////////////////////////////////////////
struct _EngineMonitor {
  double xPadding;
  double xWritten;
  double xStart;
  double xTailCur;
  double xTailMax;
  double xTailMin;
};

void EngineMonitorReset(EngineMonitor *pMonitor);

///////////////////////////////////////////////////////////////////////////////
struct _EngineParam {
  IMMDevice *pDevice;
  int bActive;
  const char *pFriendlyName;
  int bDoubleBuffered;
  AUDCLNT_SHAREMODE eShareMode;
  DWORD dwStreamFlags;
  REFERENCE_TIME hnsDevicePeriod;
  int nAlignment;
  LARGE_INTEGER DueTime;
  LONG lPeriod;
  UINT32 uNumFramesDevice;
  UINT32 uNumFramesRing;
  DWORD dwChunkSizeSource;
  WORD uFramesPlaybackBegin;
  double xFramesPlaybackBegin;
  int nMaxLatency;
  int nLastPause;
  EngineServerTime time;
  EngineServerVolume volume;
  RingCreateFun *RingCreate;
};

int EngineParamCreate(EngineParam *pParam, const EngineOptions *pOptions,
    const EngineClient *pClient);

///////////////////////////////////////////////////////////////////////////////
struct _EngineStats {
  UINT32 uNumTracks;
  UINT32 uNumFramesMax;
  UINT32 uNumFramesDropout;
} stats;

void EngineStatsReset(EngineStats *pStats);

///////////////////////////////////////////////////////////////////////////////
struct _IEngineClientState {
  const char *pId;
  int (*Enter)(EngineClient *pClient);
  void (*Leave)(EngineClient *pClient);
};

extern IEngineClientState gcaIEngineClientState[eEngineClientStateCount];

struct _EngineClient {
  Fsm fsm;
  Player *pPlayer;
  EngineServer *pServer;
  CRITICAL_SECTION cs;

  struct _EngineClientTimer {
    HANDLE hEvent;
    HANDLE hTimer;
  } timer;


  struct _EngineClientThread {
    HANDLE hEvent;
    HANDLE hThread;
    DWORD dwThreadId;
  } thread;

  struct {
    HANDLE hThread;
    int nPrevious;
    int bAltered;
  } winamp;

  struct {
    int nPrevious;
    int bAltered;
  } drain;

  EngineOptions options;
  EngineWfx wfx;
  EngineParam param;
  EngineMonitor monitor;
  EngineStats stats;
  Ring ring;
  IAudioClient *pAudioClient;
};

int EngineClientCreate(EngineClient *pClient, Player *pPlayer);
void EngineClientDestroy(EngineClient *pClient);

int EngineClientRequestServerState(EngineClient *pClient,
    EEngineServer eRequest, const char *file, int line, const char *func);

int EngineClientOpen(EngineClient *pClient, int samplerate, int numchannels,
    int bitspersamp, int *pnMaxLatency, int *pnFramesDevice);
void EngineClientClose(EngineClient *pClient);

int EngineClientCreateRing(EngineClient *pClient, RingCreateFun *RingCreate,
    DWORD dwChunkSizeSource);
int EngineClientAvailable(EngineClient *pClient);
int EngineClientWrite(EngineClient *pClient, BYTE *buf, DWORD dwChunkSize);
int EngineClientIsPlaying(EngineClient *pClient);
int EngineClientFlush(EngineClient *pClient, int ms);
int EngineClientPause(EngineClient *pClient, int nPause);
int EngineClientOutputTime(EngineClient *pClient);
int EngineClientWrittenTime(EngineClient *pClient);
void EngineClientVolume(EngineClient *pClient, Player *pPlayer);
int EngineClientApiWrite(EngineClient *pClient, int msTimeout, const char *buf,
    DWORD dwChunkSize);
int EngineClientApiEndOfTrack(EngineClient *pClient, int msTimeout);
void EngineClientApiKill(EngineClient *pClient);

///////////////////////////////////////////////////////////////////////////////
enum _EPresentServerState {
  ePresentServerStateStart,
#define YASAPI_PRESENT_SERVER_STATE(x) ePresentServerState ## x, // [
#include <yasapi_present_server_state.h>
#undef YASAPI_PRESENT_SERVER_STATE // ]
  ePresentServerStateMax,
  ePresentServerStateCount=ePresentServerStateMax-1,
  ePresentServerStateBase=ePresentServerStateTimer,
};

////////////////
struct _IPresentServerState {
  const char *pId;
  int (*Enter)(PresentServer *pPresentServer);
  void (*Leave)(PresentServer *pPresentServer);
};

extern IPresentServerState gcaIPresentServerState[ePresentServerStateCount];

////////////////
struct _PresentServer {
  Fsm fsm;
  PresentClient *pClient;
  HANDLE hEvent;
  HANDLE hTimer;
  EPresentServerState eStateRequest;
};

DWORD WINAPI PresentServerThreadProc(LPVOID lParameter);

////////////////
enum _EPresentClientState {
  ePresentClientStateStart,
#define YASAPI_PRESENT_CLIENT_STATE(x) ePresentClientState ## x, // [
#include <yasapi_present_client_state.h>
#undef YASAPI_PRESENT_CLIENT_STATE // ]
  ePresentClientStateMax,
  ePresentClientStateCount=ePresentClientStateMax-1,
  ePresentClientStateBase=ePresentClientStateThread,
};

////////////////
struct _IPresentClientState {
  const char *pId;
  int (*Enter)(PresentClient *pPresentClient);
  void (*Leave)(PresentClient *pPresentClient);
};

extern IPresentClientState gcaIPresentClientState[ePresentClientStateCount];

////////////////
struct _PresentClient {
  Fsm fsm;
  Player *pPlayer;
  HANDLE hEvent;
  HANDLE hThread;
  DWORD dwThreadId;
  PresentServer *pServer;
};

int PresentClientCreate(PresentClient *pClient, Player *Player);
void PresentClientDestroy(PresentClient *pClient);

///////////////////////////////////////////////////////////////////////////////
enum _EGtkServer {
  eGtkServerStateStart,
#define YASAPI_GTK_SERVER(x) eGtkServerState ## x, // [
#include <yasapi_gtk_server_state.h>
#undef YASAPI_GTK_SERVER// ]
  eGtkServerStateMax,
  eGtkServerStateBase=eGtkServerStateThreadEvent,
  eGtkServerStateReady=eGtkServerStateMax-1,
  eGtkServerStateCount=eGtkServerStateMax-1,
};

struct _IGtkServerState {
  const char *pId;
  int (*Enter)(GtkServer *pServer);
  void (*Leave)(GtkServer *pServer);
};

extern IGtkServerState gcaIGtkServerState[eGtkServerStateCount];

///////////////////////////////////////////////////////////////////////////////
enum _EGtkLib {
  eGtkLibGlib,
  eGtkLibGObject,
  eGtkLibGdk,
  eGtkLibGtk,
  eGtkLibCount,
};

struct _GtkServer {
  Fsm fsm;
  GtkClient *pClient;
  OptionsMemberGtk options;
  HMODULE ahGtkLib[eGtkLibCount];
  HMODULE *phGtkLib;

  struct {
    HANDLE hEvent;
    EGtkServer eStateRequest;
  } thread;
};

DWORD WINAPI GtkServerThreadProc(LPVOID lParameter);

///////////////////////////////////////////////////////////////////////////////
enum _EGtkClient {
  eGtkClientStateStart,
#define YASAPI_GTK_CLIENT(x) eGtkClientState ## x, // [
#include <yasapi_gtk_client_state.h>
#undef YASAPI_GTK_CLIENT // ]
  eGtkClientStateMax,
  eGtkClientStateCount=eGtkClientStateMax-1,
};

struct _IGtkClientState {
  const char *pId;
  int (*Enter)(GtkClient *pClient);
  void (*Leave)(GtkClient *pClient);
};

extern IGtkClientState gcaIGtkClientState[eGtkClientStateCount];

///////////////////////////////////////////////////////////////////////////////
enum _EPlayerState {
  ePlayerStateStart,
#define YASAPI_PLAYER_STATE(x) ePlayerState ## x, // [
#include <yasapi_player_state.h>
#undef YASAPI_PLAYER_STATE // ]
  ePlayerStateMax,
  ePlayerStateCount=ePlayerStateMax-1,
};

struct _IPlayerState {
  const char *pId;
  int (*Enter)(Player *pPlayer);
  void (*Leave)(Player *pPlayer);
};

extern IPlayerState gcaIPlayerState[ePlayerStateCount];

enum _EPlayerLimits {
  ePlayerVolumeMax=255,
};

struct _Player {
  Fsm fsm;
  CRITICAL_SECTION cs;
  int nLockRecursion;
  Out_Module *pOut;
  wchar_t path[MAX_PATH];
  const wchar_t *pBasename;
  wchar_t ini[MAX_PATH];

  struct {
    int nVolume;
  } param;

  struct {
    UINT32 uNumTracks;
    UINT32 uNumGapless;
    double xMsDropout;
    UINT32 uMsLatencyMax;
    UINT32 uNumGaplessBroken;
  } stats;

  Cache cache;
  PresentClient present;
  EngineClient engine;
  GtkClient gtk;
};

int PlayerCreate(Player *pPlayer, Out_Module *pOut);
void PlayerDestroy(Player *pPlayer);

int PlayerLock(Player *pPlayer );
void PlayerUnlock(Player *pPlayer);
int PlayerLockDeep(Player *pPlayer, int nLockRecursion );
void PlayerUnlockDeep(Player *pPlayer, int *pnLockRecursion);

int PlayerPropertiesPath(Player *pPlayer);

int PlayerOpen(Player *pPlayer, int samplerate, int numchannels,
    int bitspersamp, int *pnMaxLatency, int *pnFramesDevice);
void PlayerClose(Player *pPlayer);

void PlayerFlush(Player *pPlayer, int ms);
int PlayerPause(Player *pPlayer, int nPause);
int PlayerIsPlaying(Player *pPlayer);
int PlayerAvailable(Player *pPlayer);
int PlayerOutputTime(Player *pPlayer);
int PlayerWrittenTime(Player *pPlayer);
int PlayerWrite(Player *pPlayer, BYTE *buf, int len);
void PlayerVolume(Player *pPlayer, int nVolume);
void PlayerSetActive(Player *pPlayer, int bActive);

const OptionsCommon *PlayerOptionsCommonConfig(Player *pPlayer);
const OptionsDevice *PlayerOptionsDeviceConfig(Player *pPlayer);

///////////////////////////////////////////////////////////////////////////////
const char *Bitdepth(EBitdepth eBitdepth);

#ifdef __cplusplus // [
}
#endif // ]
#endif // __YASAPI_H__ ]
