/*
 * yasapi_config.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi_config.h>
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
static void ConfigApply(Dialog *pDialog, int moffs, const char *op);

///////////////////////////////////////////////////////////////////////////////
const char CONFIG_WIDGET[]="ConfigWidget";

char *NavigatePlayer(Dialog *pDialog, int offs)
{
  Player *pPlayer=pDialog->pGtk->pPlayer;

  return ((char *)pPlayer)+offs;
}

char *NavigateDeviceSelectedConfig(Dialog *pDialog, int offs)
{
  Player *pPlayer=pDialog->pGtk->pPlayer;
  Cache *pCache=&pPlayer->cache;
  CacheDevice *pCacheDevice=pCache->selected.config.pCacheDevice;

  return ((char *)pCacheDevice)+offs;
}

char *NavigateOptionsCommonConfig(Dialog *pDialog, int offs)
{
  Player *pPlayer=pDialog->pGtk->pPlayer;
  Cache *pCache=&pPlayer->cache;
  OptionsCommon *pOptionsCommon=&pCache->options.common.config;

  return ((char *)pOptionsCommon)+offs;
}

char *NavigateOptionsDeviceConfig(Dialog *pDialog, int offs)
{
  Player *pPlayer=pDialog->pGtk->pPlayer;
  Cache *pCache=&pPlayer->cache;
  CacheDevice *pCacheDevice=pCache->selected.config.pCacheDevice;
  OptionsDevice *pOptionsDevice=&pCacheDevice->pMember->config;

  return ((char *)pOptionsDevice)+offs;
}

char *NavigateEngineClient(Dialog *pDialog, int offs)
{
  EngineClient *pClient=&pDialog->pGtk->pPlayer->engine;

  return -1<FsmState(pClient)?((char *)pClient+offs):NULL;
}

char *NavigateEngineServer(Dialog *pDialog, int offs)
{
  EngineClient *pClient=&pDialog->pGtk->pPlayer->engine;
  EngineServer *pServer=pClient->pServer;

  return eEngineClientStateBase<=FsmState(pClient)&&-1<FsmState(pServer)?((char *)pServer+offs):NULL;
}

char *NavigateEngineClientMonitor(Dialog *pDialog, int offs)
{
  EngineClient *pClient=&pDialog->pGtk->pPlayer->engine;

  return -1<FsmState(pClient)?((char *)&pClient->monitor+offs):NULL;
}

char *NavigateConfig(Dialog *pDialog, int offs)
{
  return (char *)&pDialog->config+offs;
}

///////////////////////////////////////////////////////////////////////////////
ConfigWidget *ConfigWidgetCreate(ClassCell *pCell, Dialog *pDialog,
    IConfigWidgetVtbl *lpVtbl, GtkWidget *widget)
{
  ConfigWidget *pWidget;

  if (0<=pCell->widget&&lpVtbl) {
    pWidget = (ConfigWidget *)((char *)&pDialog->config + pCell->widget);
    pWidget->lpVtbl=lpVtbl;
    pWidget->pDialog=pDialog;
    pWidget->pCell=pCell;
    pWidget->widget=widget;
    g_object_set_data(G_OBJECT(widget), CONFIG_WIDGET, pWidget);
    // It's not a good idea to call "set()" on our individual widget
    // because it may have side effects to other, not yet initialized
    // widgets.
    // That's the theory. In practice we need both.
    lpVtbl->set(pWidget);

  pWidget->provider=NULL;
  }
  else
    pWidget = NULL;

  if (pCell->tooltip)
    gtk_widget_set_tooltip_markup (widget, pCell->tooltip);

  return pWidget;
}

OptionsCommon *DialogOptionsCommonConfig(Dialog *pDialog)
{
  static OptionsCommon options;
  Player *pPlayer=pDialog->pGtk->pPlayer;
  Cache *pCache=&pPlayer->cache;

  return pCache?&pCache->options.common.config:&options;
}

OptionsDevice *DialogOptionsDeviceConfig(Dialog *pDialog)
{
  static OptionsDevice options;
  Player *pPlayer=pDialog->pGtk->pPlayer;
  Cache *pCache=&pPlayer->cache;

  return pCache?&pCache->selected.config.pCacheDevice->pMember->config:&options;
}

static void SideEffectePresentPeriod(ConfigWidget *pWidget)
{
  Dialog *pDialog=pWidget->pDialog;
  OptionsCommon *pOptionsCommon=DialogOptionsCommonConfig(pDialog);

  gtk_widget_set_sensitive (pDialog->config.common.present.period.widget,
      pOptionsCommon->present.bEnabled);
}



static void SideEffectMode(ConfigWidget *pWidget)
{
  Dialog *pDialog=pWidget->pDialog;
  OptionsDevice *pOptionsDevice=DialogOptionsDeviceConfig(pDialog);
  int bSensitive;

  switch (pOptionsDevice->eMode) {
  case eModeShared:
    pOptionsDevice->ePeriod=ePeriodDefault;
    bSensitive=1;
    break;
  case eModeExclusive:
    pOptionsDevice->ePeriod=ePeriodMinimum;
    bSensitive=0;
    break;
  default:
    bSensitive=0;
    break;
  }

  gtk_widget_set_sensitive (pDialog->config.wasapi.auto_convert_pcm.widget,
      bSensitive);
  gtk_widget_set_sensitive (pDialog->config.wasapi.src_default_quality.widget,
      bSensitive);
}

static void SideEffectStrategy(ConfigWidget *pWidget)
{
  Dialog *pDialog=pWidget->pDialog;
  OptionsDevice *pOptionsDevice=DialogOptionsDeviceConfig(pDialog);
  int bSensitive=eStrategyPush==pOptionsDevice->eStrategy;

  gtk_widget_set_sensitive (pDialog->config.wasapi.timer_periodic.widget,
      bSensitive);
}

static void SideEffectGapless(ConfigWidget *pWidget)
{
  Dialog *pDialog=pWidget->pDialog;
  OptionsCommon *pOptionsCommon=DialogOptionsCommonConfig(pDialog);
  int bSensitive=pOptionsCommon->bGapless;

  gtk_widget_set_sensitive (pDialog->config.common.drain.priority.widget,bSensitive);
  gtk_widget_set_sensitive (pDialog->config.common.winamp.priority.widget,bSensitive);
}

static void ConfigWindowCreate(Dialog *pDialog)
{
  /////////////////////////////////////////////////////////////////////////////
  static ConfigComboBoxMap mode_map[]={
    { "Shared", "<i>Shared</i>", eModeShared },
    { "Exclusive", "<i>Exclusive</i>", eModeExclusive },
    { NULL, NULL, -1 }
  };

  static ConfigComboBoxMap strategy_map[]={
    { "Push", "<i>Push</i>", eStrategyPush },
    { "Pull", "<i>Pull</i>", eStrategyPull },
    { NULL, NULL, -1 }
  };

  static const char attDevice[]="Choose a <i>WASAPI</i> enabled device.";

  static const char attMode[]=
    "Whether the plug-in should operate with <i>WASAPI</i> mode"
      " <i>Exclusive</i> or <i>Shared</i>."
    ;
  static const char attStrategy[]=
    "Whether the plug-in should operate with <i>WASAPI</i> strategy"
      " <i>Push</i> or <i>Pull</i>."
    ;

  static ClassPage general_rows={
    { LABEL_STATIC(1,0,"Device:",attDevice),
        COMBO_BOX_DEVICE_WASAPI(1,0,general.device,attDevice) },
    { LABEL_STATIC(1,0,"Mode:",attMode), COMBO_BOX_DEVICE(1,0,
        general.mode,eMode,mode_map,SideEffectMode,attMode) },
    { LABEL_STATIC(1,0,"Strategy:",attStrategy), COMBO_BOX_DEVICE(1,0,
        general.strategy,eStrategy,strategy_map,SideEffectStrategy,
        attStrategy) },
    { { NULL } }
  };

  ////////////////
  static const char attLatency[]=
      "Whether <i>YASAPI/NT</i> should optimize getoutputtime() and"
      " getwrittentime() by not using IAudioClock and always return zero"
      " and the latency resulting from <i>YASAPI/NT</i>'s buffer,"
      " respectively."
      ;
  static const char attMono2Stereo[]=
      "Duplicate the only channel from mono to a second channel for (pseudo)"
      " stereo (2.0)."
      ;
  static const char attVolume[]=
      "Enable volume control."
      ;
  static const char attDither[]=
      "Enable rounding/dithering for"
      " volume control."
      ;
  static const char attSurround[]="How <i>4</i>, <i>6</i>, and <i>8</i>"
      " channels shoud be interpreted: as <i>surround</i>,"
      " <i>5.1 surround</i>, and <i>7.1 surround</i> (enabled),"
      " or as <i>quadrophonic</i>, <i>5.1</i>, and <i>7.1</i> (disabled)."
      ;
  static const char attDisconnect[]="Wether <i>YASAPI/NT</i> should"
      " release it's instances of IAudioClient, IAudioRenderClient, and"
      " IAudioClock during pause and re-initialize them when done."
      ;
  static const char attGapless[]=
      "Enable gapless playback."
      ;
  static const char attThemePadding[]=
      "The color theme for the progress bar <i>#Frames Padding</i>"
      " on page <i>Monitor</i>."
      ;
  static const char attThemeStart[]=
      "The color theme for the progress bar <i>Start Playing</i>"
      " on page <i>Monitor</i>."
      ;
  static const char attThemeEngineTail[]=
      "The color theme for the progress bars of group"
      " <i>Max./Prev./Min. Gap</i> on page <i>Monitor</i>."
      ;
  static const char attThemeWritten[]=
      "The color theme for the progress bar <i>#Frames Written</i>"
      " on page <i>Monitor</i>."
      ;
  static const char attPriorityDrain[]=
    "Increase the priority of <i>YASAPI/NT</i>'s render thread when draining."
    ;
  static const char attPriorityWinamp[]=
    "Decrease the priority of <i>Winamp</i>'s thread when draining."
    ;

  static ConfigComboBoxMap mono2stero_map[]={
    { "Off","<i>Off",eMono2StereoOff },
    { "Duplicate Plain","<i>Duplicate Plain</i>",eMono2StereoPlain },
    { "Duplicate Mutiplied by ½","<i>Duplicate Mutiplied by ½</i>",
        eMono2StereoHalf },
    { "Duplicate Mutiplied by 1/√2","<i>Duplicate Mutiplied by 1/√2</i>",
        eMono2StereoSqrt2Inverse },
    { NULL, NULL, -1 }
  };


  static ConfigComboBoxMap volume_map[]={
    { "Off","<i>Off</i>",eVolumeOff },
    { "YASAP/NT","<i>YASAP/NT</i>",eVolumeSoftware },
    { "WASAPI","<i>WASAPI</i>",eVolumeWasapi },
    { NULL, NULL, -1 }
  };

  static ConfigComboBoxMap dither_map[]={
    { "Off/Truncate","<i>Off/Truncate</i>",eDitherTruncate },
    { "Round","<i>Round</i>",eDitherRound },
    { "Simple Dither","<i>Simple Dither</i>",eDitherSimple },
    { NULL, NULL, -1 }
  };

  static ConfigComboBoxMap theme_map[]={
    { "Default","<i>Default</i>",eConfigThemeDefault },
    { "Red","<i>Red</i>",eConfigThemeRed },
    { "Green","<i>Green</i>",eConfigThemeGreen },
    { "Blue","<i>Blue</i>",eConfigThemeBlue },
    { NULL, NULL, -1 }
  };

  static ConfigComboBoxMap theme_map_min[]={
    { "Default","<i>Default</i>",eConfigThemeDefault },
    { "Red","<i>Red</i>",eConfigThemeRed },
    { "Green","<i>Green</i>",eConfigThemeGreen },
    { NULL, NULL, -1 }
  };

  static ConfigComboBoxMap render_priority_map[]={
    { "Normal", "<i>Normal</i>", eThreadPriorityNormal },
    { "Above Normal", "<i>Above Normal</i>", eThreadPriorityAboveNormal },
    { "Highest", "<i>Highest</i>", eThreadPriorityHighest },
    { "Time Critical", "<i>Time Critical</i>", eThreadPriorityTimeCritical },
    { NULL }
  };

  static ConfigComboBoxMap winamp_prority_map[]={
    { "Normal", "<i>Normal</i>", eThreadPriorityNormal },
    { "Below Normal", "<i>Below Normal</i>", eThreadPriorityBelowNormal },
    { "Lowest", "<i>Lowest</i>", eThreadPriorityLowest },
    { "Idle", "<i>Idle</i>", eThreadPriorityIdle },
    { NULL }
  };

  static ConfigComboBoxMap drain_prority_map[]={
    { "Normal", "<i>Normal</i>", eThreadPriorityNormal },
    { "Above Normal", "<i>Above Normal</i>", eThreadPriorityAboveNormal },
    { "Highest", "<i>Highest</i>", eThreadPriorityHighest },
    { "Time Critical", "<i>Time Critical</i>", eThreadPriorityTimeCritical },
    { NULL }
  };

  static const char attPresent[]="Wether <i>YASAPI/NT</i> should"
      " push the <i>Configuration</i> and <i>About</i> dialogs in"
      " front when the visualization is running."
      ;
//  static const char attPresentPeriod[]="Wether <i>YASAPI/NT</i> should"
//      " push the <i>Configuration</i> and <i>About</i> dialogs in"
//      " front when the visualization is running."
//      ;
  static ConfigScaleParameter present_period_param={
    .min=100.0,
    .max=10000.0,
    .step=100.0,
  };

  static ClassPage present_rows_frame={
    {  CHECK_BUTTON_COMMON_SE(1,0,common.present.enabled,present.bEnabled,
        SideEffectePresentPeriod,attPresent),
        LABEL_STATIC(1,0,"each",NULL),
        SCALE_COMMON(1,0,common.present.period,present.xPeriod,
          &present_period_param,NULL),
        LABEL_STATIC(1,0,"ms",NULL) },
    { { NULL } }
  };

  static ClassPage render_frame={
    { COMBO_BOX_COMMON(1,0,common.drain.priority,drain.ePriority,drain_prority_map,
      NULL,attPriorityDrain),
      LABEL_STATIC_WEIGHT(1,0,"Priority of <i>YASAPI/NT</i>'s Render Thread when <i>Draining</i>",attPriorityDrain) },
    { { NULL } }
  };

  static ClassPage background_frame={
    { COMBO_BOX_COMMON(1,0,common.winamp.priority,winamp.ePriority,winamp_prority_map,
      NULL,attPriorityWinamp),
        LABEL_STATIC_WEIGHT(1,0,"Priority of <i>Winamp</i>'s Thread when <i>Draining</i>",attPriorityWinamp) },
    { { NULL } }
  };

  static ClassPage common_misc_rows={
    { LABEL_STATIC(1,0,"Report Latency Instead of Time:",attLatency), 
        CHECK_BUTTON_COMMON(3,0,common.latency,bLatency,attLatency) },
    { LABEL_STATIC(1,0,"Play Mono as Stereo:",attMono2Stereo), 
        COMBO_BOX_COMMON(3,0,common.mono2stereo,eMono2Stereo,
            mono2stero_map,NULL,attMono2Stereo) },
    { LABEL_STATIC(1,0,"Volume Control:",attVolume), 
        COMBO_BOX_COMMON(3,0,common.volume,eVolume,volume_map,NULL,attVolume) },
    { LABEL_STATIC(1,0,"Rounding/Dithering:",attDither), 
        COMBO_BOX_COMMON(3,0,common.dither,eDither,dither_map,NULL,attDither) },
    { LABEL_STATIC(1,0,"Surround:",attSurround), 
        CHECK_BUTTON_COMMON(3,0,common.surround,bSurround,attSurround) },
    { LABEL_STATIC(1,0,"Disconnect when Pause:",attDisconnect), 
        CHECK_BUTTON_COMMON(3,0,common.disconnect,bDisconnect,attDisconnect) },
    { LABEL_STATIC(1,0,"Gapless Playback:",attGapless), 
        CHECK_BUTTON_COMMON_SE(3,0,common.gapless,bGapless,SideEffectGapless,attGapless) },
    { { NULL },
        FRAME(1,render_frame) },
    { { NULL },
        FRAME(1,background_frame) },
    { LABEL_STATIC(1,0,"Present Dialogs:",attPresent),
        FRAME(2,present_rows_frame) },
    { { NULL } }
  };

  static ClassPage common_colors_rows={
    { LABEL_STATIC(1,0,"Theme for Monitoring <i>#Frames Padding</i>:",
            attThemePadding), 
        COMBO_BOX_COMMON(2,0,common.theme.padding,eThemePadding,
            theme_map_min,ConfigMonitorPaddingSetStyle,attThemePadding) },
    { LABEL_STATIC(1,0,"Theme for Monitoring <i>#Frames Start Playing</i>:"
            ,attThemeStart), 
        COMBO_BOX_COMMON(2,0,common.theme.start,eThemeStart,
            theme_map_min,ConfigMonitorStartSetStyle,attThemeStart) },
    { LABEL_STATIC(1,0,"Theme for Monitoring <i>Gap Group</i>:",
        attThemeEngineTail), 
        COMBO_BOX_COMMON(2,0,common.theme.engine_tail,eThemeEngineTail,
            theme_map,ConfigMonitorEngineClientTailSetStyle,attThemeEngineTail) },
    { LABEL_STATIC(1,0,"Theme for Monitoring <i>#Frames Written</i>:",
            attThemeWritten), 
        COMBO_BOX_COMMON(2,0,common.theme.written,eThemeWritten,
            theme_map_min,ConfigMonitorWrittenSetStyle,attThemeWritten) },
    { { NULL } }
  };

  static ClassNotebook notebook_common_pages={
    { LABEL_STATIC(1,0,"Miscellaneous",NULL),PAGE(common_misc_rows) },
    { LABEL_STATIC(1,0,"Colors",NULL),PAGE(common_colors_rows) },
    { { NULL } }
  };

  ////////////////
  static ConfigComboBoxMap period_map[]={
    { "Default", "<i>Default</i>", ePeriodDefault },
    { "Minimum", "<i>Minimum</i>", ePeriodMinimum },
    { NULL }
  };

  static const char attAutoConvertPCM[]=
      "Whether <i>WASAPI</i> should automatically convert to the desiered"
      " format. Available only with mode <i>Shared</i>. Disabled otherwise."
      ;
  static const char attSRCDefaultQuality[]=
      "Whether (according to <i>Microsoft</i>) \"a sample rate converter with"
      " better quality than the default conversion but with a higher"
      " performance cost is used. This should be used if the audio is"
      " ultimately intended to be heard by humans as opposed to other"
      " scenarios such as pumping silence or populating a meter.\""
      ;
  static const char attTimerPeriodic[]=
    "<i>Microsoft</i> waitable timers are not strictly <i>WASAPI</i>"
    " objects. <i>YASAPI/NT</i> uses one in order to implement strategy"
    " <i>Push</i> and the option fits perfectly on this page. If enabled"
    " <i>YASAPI/NT</i> starts the timer once initialized with an appropriate"
    " period matching a playback cycle. Otherwise the timer ist re-started"
    " again with each playback cycle."
    ;
  static const char attDevicePeriod[]=
    "Associated with <i>WASAPI</i> mode. According"
    " to <i>Microsoft</i> \"[<i>Default</i>] specifies the [...] period"
    " for a shared-mode stream. [<i>Minimum</i>] specifies the [...]"
    " period for an exclusive-mode stream.\""
    ;
  static const char attPriorityRender[]=
    "Increase the priority of <i>YASAPI/NT</i>'s render thread"
    " in order to minimize disturbions."
    ;
  static const char attProAudio[]=
    "Associate <i>YASAPI/NT</i>'s render thread with the \"Pro Audio\""
    " task. Should at least be enabled with mode <i>Exclusive</i> and"
    " strategy <i>Pull</i>."
    ;
  static const char attAlignmentDance[]=
    "Do the \"alignment dance\" when error AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED"
    " occurs with IAudioClient::Initialize()."
    ;

  static ClassPage wasapi_rows_frame={
    { LABEL_STATIC(1,0,"Mode:",attMode),LABEL_MAPPED_DEVICE(2,0,
        wasapi.mode,eMode,mode_map,attMode) },
    { { NULL },
        CHECK_BUTTON_DEVICE(1,0,wasapi.auto_convert_pcm,bAutoConvertPCM,
            attAutoConvertPCM),
        LABEL_STATIC(1,0,"Auto Convert PCM",attAutoConvertPCM) },
    { { NULL },
        CHECK_BUTTON_DEVICE(1,0,wasapi.src_default_quality,bSRCDefaultQuality,
            attSRCDefaultQuality),
        LABEL_STATIC(1,0,"SRC Default Quality",
            attSRCDefaultQuality) },
    { LABEL_STATIC(1,0,"Device Period:",attDevicePeriod),
        LABEL_MAPPED_DEVICE(2,0,wasapi.period,ePeriod,period_map,
            attDevicePeriod) },
    { LABEL_STATIC(1,0,"Strategy:",attStrategy),
        LABEL_MAPPED_DEVICE(2,0,wasapi.strategy,eStrategy,strategy_map,
            attStrategy) },
    { { NULL },
        CHECK_BUTTON_DEVICE(1,0,wasapi.timer_periodic,bTimerPeriodic,
            attTimerPeriodic),
        LABEL_STATIC(1,0,"Periodic Timer",attTimerPeriodic) },
    { LABEL_STATIC(1,0,"Priority of Render Thread:",attPriorityRender), 
        COMBO_BOX_DEVICE(2,0,wasapi.render.priority,render.ePriority,
            render_priority_map,NULL,attPriorityRender) },
    { LABEL_STATIC(1,0,"Associate Render Thread with \"Pro Audio\" Task:",
        attProAudio), CHECK_BUTTON_DEVICE(2,0,wasapi.pro_audio,bProAudio,
            attProAudio) },
    { LABEL_STATIC(1,0,"Alignment Dance:",attAlignmentDance),
        CHECK_BUTTON_DEVICE(2,0,wasapi.alignment_dance,bAlignmentDance,
            attAlignmentDance) },
    { { NULL } }
  };

  static ClassPage wasapi_rows={
    { LABEL_DEVICE_WASAPI(1,0,wasapi.device) },
    { FRAME(1,wasapi_rows_frame) },
    { { NULL } }
  };

  ////////////////
  static ConfigComboBoxMap bitdepth_map[]={
    { "As Delivered by Winamp", "<i>As Delivered by Winamp</i>",
        eBitdepthDisabled },
    { "16 Bit", "<i>16 Bit</i>", eBitdepth16 },
    { "24 Bit", "<i>24 Bit</i>", eBitdepth24 },
    { "32 Bit", "<i>32 Bit</i>", eBitdepth32 },
    { NULL, NULL, -1 }
  };


  static ConfigScaleParameter latency_param={
    .min=0.0,
    .max=1000.0,
    .step=10.0,
  };

  static ConfigScaleParameter playback_begin_param={
    .min=1.0,
    //.max=10.0,
    .max=20.0,
    .step=0.1,
  };


  static const char attBitdepth[]=
    "Whether <i>YASAPI/NT</i> should convert to the choosen format."
    ;
  static const char attPlaybackBegin[]=
    "The minimum number of audio frames which should be cached by"
    " <i>YASAPI/NT</i> before starting to output."
    ;
  static const char attMaxLatency[]=
    "When <i>Winamp</i> switches to a new track it requests the output"
    " plug-in to report the maximum latency. <i>YASAPI/NT</i> is"
    " not able to do so before knowing the size of audio chunks"
    " <i>Winamp</i> or the current input plug-in, respectively, is"
    " going to send. This option is meant to resolve the dilemma: Have"
    " an eye at <i>Configure → Monitor</i> and after having run all your"
    " DSP/input plug-ins set this option accordingly."
    ;

  static ClassPage yasapi_rows_frame={
    { LABEL_STATIC(1,0,"Bit Depth:",attBitdepth),
        COMBO_BOX_DEVICE(1,0,yasapi.bitdepth,eBitdepth,bitdepth_map,
            NULL,attBitdepth) },
    { LABEL_STATIC(1,0,"Max. Latency:",attMaxLatency),
        SCALE_DEVICE(2,0,yasapi.max_latency_assumed,xMaxLatencyAssumed,
            &latency_param,attMaxLatency),
        LABEL_STATIC(3,0,"ms",attMaxLatency)
    },
    { LABEL_STATIC(1,0,"Start Playing:",attPlaybackBegin),
        SCALE_DEVICE(2,0,yasapi.playback_begin,xPlaybackBegin,
            &playback_begin_param,attPlaybackBegin),
        LABEL_STATIC(1,0,"×",NULL),
        LABEL_MAPPED_DEVICE(1,0,yasapi.period_playback_begin,ePeriod,
            period_map,attDevicePeriod),
        LABEL_STATIC(1,0,"Buffer Size",NULL) },
    { { NULL } }
  };

  static ClassPage yasapi_rows={
    { LABEL_DEVICE_WASAPI(1,0,yasapi.device) },
    { FRAME(1,yasapi_rows_frame) },
    { { NULL } }
  };

  ////////////////
  static ConfigProgessBarParameter param_monitor_padding={ 0 };
  static ConfigProgessBarParameter param_monitor_written={ 0 };
  static ConfigProgessBarParameter param_monitor_start={ 0 };
  static ConfigProgessBarParameter param_monitor_tail_min={ .height=6 };
  static ConfigProgessBarParameter param_monitor_tail_cur={ .height=6 };
  static ConfigProgessBarParameter param_monitor_tail_max={ .height=6 };

  static const char attWinampNumTracks[]=
    "Within the current <i>Winamp</i> session and any device,"
    " the overall number of individual tracks played."
    ;
  static const char attWinampNumGaplessSequences[]=
    "Within the current <i>Winamp</i> session and any device,"
    " the overall number of gapless track sequences played."
    " A gapless track sequence may"
    " in turn consist of several individual tracks sharing the same format."
    " Tracks of different format cannot played gaplessly"
    ;
  static const char attWinampDropouts[]=
    "Within the current <i>Winamp</i> session and any device,"
    " the overall number of dropouts."
    ;
  static const char attWinampGaplessBroken[]=
    "Within the current <i>Winamp</i> session and any device, the overall"
    " number of tacks reacheing EOT (end of track) before being able to"
    " continue gaplessly."
    ;
  static const char attWinampMaxLatency[]=
    "The latency accumulated so far with the current <i>Winamp</i> session"
    " and any device."
    ;
  static const char attPadding[]=
    "Taken at the beginning of a playback cycle, the number of not yet"
    " played audio frames cached in the <i>WASAPI</i> device buffer."
    " (Not defined in case of double buffering, i.e. exclusive pulled"
    " playback. Remains empty in this case.)"
    ;
  static const char attWritten[]=
    "Taken at the beginning of a playback cycle, the number of not yet"
    " played audio frames cached in <i>YASAPI/NT</i>'s buffer."
    ;
  static const char attStart[]=
    "The minimum number of audio frames needed to be written to"
    " <i>YASAPI/NT</i>'s buffer before it starts to play (cf. page"
    " <i>YASAPI/NT</i>.)"
    ;
  static const char attTail[]=
    "The minimum, the current, and the maximum number of audio frames so"
    " far left in <i>YASAPI/NT</i>'s buffer, respectively, after having"
    " bridged the gap between tracks of the current sequence."
    " Note: The gap is right to the bar displaying the number of frames"
    " left in the buffer."
    ;
  static const char attTailMax[]=
    "The minimum number of frames so far left in <i>YASAPI/NT</i>'s buffer"
    " after having bridged the gap between tracks of the current sequence."
    " Note: The gap is right to the bar displaying the number of frames"
    " left in the buffer."
    ;
  static const char attTailMin[]=
    "The maximum number of frames so far left in <i>YASAPI/NT</i>'s buffer"
    " after having bridged the gap between tracks of the current sequence."
    " Note: The gap is right to the bar displaying the number of frames"
    " left in the buffer."
    ;
  static const char attTailCur[]=
    "The number of frames left in <i>YASAPI/NT</i>'s buffer after having"
    " bridged the gap to the previous track."
    " Note: The gap is right to the bar displaying the number of frames"
    " left in the buffer."
    ;
  static const char attCurDevice[]=
    "The device in use."
    ;
  static const char attTracksEngine[]=
    "Within the current gapless sequence, the number of tracks played."
    ;
  static const char attFramesDevice[]=
    "The size of the buffer shared with the <i>WASAPI</i> device."
    ;
  static const char attFramesYasapi[]=
    "The size of <i>YASAPI/NT</i>'s buffer."
    ;
  static const char attEngineDropouts[]=
    "Within the current gapless sequence, the number of dropouts."
    ;
  static const char attLatencyGapless[]=
    "The latency accumulated so far with the current gapless track sequence."
    ;

  static ClassPage monitor_tail_rows={
    { LEVEL_BAR_ENGINE(1,0,&param_monitor_tail_min,monitor.engine.tail_min,
        monitor.xTailMin,attTailMax) },
    { LEVEL_BAR_ENGINE(1,0,&param_monitor_tail_cur,monitor.engine.tail_cur,
        monitor.xTailCur,attTailCur) },
    { LEVEL_BAR_ENGINE(1,0,&param_monitor_tail_max,monitor.engine.tail_max,
        monitor.xTailMax,attTailMin) },
    { { NULL } }
  };

  static ClassPage monitor_misc_rows={
    // {
    { LABEL_STATIC(1,0,"<b>Winamp Session</b>",NULL) },
    { LABEL_STATIC(1,0,"#Tracks:",attWinampNumTracks),
        LABEL_UINT32_PLAYER(1,0,
        "%u",monitor.player.num_tracks,stats.uNumTracks,attWinampNumTracks) },
    { LABEL_STATIC(1,0,"#Gapless Sequences:",attWinampNumGaplessSequences),
        LABEL_UINT32_PLAYER(1,0,
        "%u",monitor.player.num_gapless,stats.uNumGapless,
            attWinampNumGaplessSequences) },
    { LABEL_STATIC(1,0,"#Broken Sequences:",attWinampGaplessBroken),
        LABEL_UINT32_PLAYER(1,0,"%u",monitor.player.num_gapless_broken,
            stats.uNumGaplessBroken,attWinampGaplessBroken) },
    { LABEL_STATIC(1,0,"Max. Latency:",attWinampMaxLatency),
        LABEL_UINT32_PLAYER(1,0,"%u ms",monitor.player.ms_latency,
            stats.uMsLatencyMax,attWinampMaxLatency) },
    { LABEL_STATIC(1,0,"Dropout:",attWinampDropouts),
        LABEL_DOUBLE_PLAYER(1,0,"%.0f ms",monitor.player.ms_dropout,
            stats.xMsDropout,attWinampDropouts) },
    // }
    { LABEL_STATIC(1,0,"",NULL) },
    // {
    { LABEL_STATIC(1,0,"<b>Engine State</b>",NULL) },
    { LABEL_STATIC(1,0,"Client:",NULL),
        LABEL_ENGINE(1,0,monitor.state.client,fsm.pId,NULL) },
    { LABEL_STATIC(1,0,"Server:",NULL),
        LABEL_ENGINE_SERVER(1,0,monitor.state.server,fsm.pId,NULL) },
    // }
    { LABEL_STATIC(1,0,"",NULL) },
    // {
    { LABEL_STATIC(1,0,"<b>Current Track"
        " / Current Gapless Track Sequence</b>",NULL) },
    { LABEL_STATIC(1,0,"Device:",attCurDevice),
        LABEL_ENGINE(1,0,monitor.device,param.pFriendlyName,attCurDevice) },
    { LABEL_STATIC(1,0,"Volume Control:",NULL),
        LABEL_VOLUME(1,0,monitor.engine.volume,NULL) },
    { LABEL_STATIC(1,0,"#Tracks:",attTracksEngine), LABEL_UINT32_ENGINE(1,0,
        "%u",monitor.engine.num_tracks,stats.uNumTracks,attTracksEngine) },
    { LABEL_STATIC(1,0,"Dropout:",attEngineDropouts),
        LABEL_FRAMES_ENGINE(1,0,monitor.engine.num_frames_dropout,
            stats.uNumFramesDropout,attEngineDropouts) },
    // }
    { { NULL } }
  };

  static ClassPage monitor_buffers_rows={
    // {
    { LABEL_STATIC(1,0,"<b>WASAPI Device</b>",NULL) },
    { LABEL_STATIC(1,0,"Size:",attFramesDevice), LABEL_FRAMES_ENGINE(1,0,
        monitor.engine.num_frames_device,param.uNumFramesDevice,
            attFramesDevice) },
    { LABEL_STATIC(1,0,"#Frames Padding:",attPadding), LEVEL_BAR_ENGINE(1,0,
        &param_monitor_padding,monitor.padding,monitor.xPadding,attPadding) },
    // }
    { LABEL_STATIC(1,0,"",NULL) },
    // {
    { LABEL_STATIC(1,0,"<b>YASAPI/NT</b>",NULL) },
    { LABEL_STATIC(1,0,"Size:",attFramesYasapi), LABEL_FRAMES_ENGINE(1,0,
        monitor.engine.num_frames_ring,param.uNumFramesRing,
            attFramesYasapi) },
    { LABEL_STATIC(1,0,"Max. Latency:",attLatencyGapless),
        LABEL_FRAMES_ENGINE(1,0,monitor.engine.num_frames_latency,
            stats.uNumFramesMax,attLatencyGapless) },
    { LABEL_STATIC(1,0,"Start Playing:",attStart), LEVEL_BAR_ENGINE(1,0,
        &param_monitor_start,monitor.start,monitor.xStart,attStart) },
    { LABEL_STATIC(1,0,"Max./Prev./Min. Gap:",attTail),
        FRAME(1,monitor_tail_rows) },
    { LABEL_STATIC(1,0,"#Frames Written:",attWritten), LEVEL_BAR_ENGINE(1,0,
        &param_monitor_written,monitor.written,monitor.xWritten,attWritten) },
    // }
    { { NULL } }
  };

  static ClassNotebook notebook_monitor_pages={
    { LABEL_STATIC(1,0,"Miscellaneous",NULL),PAGE(monitor_misc_rows) },
    { LABEL_STATIC(1,0,"Buffers",NULL),PAGE(monitor_buffers_rows) },
    { { NULL } }
  };

  ////////////////
  static ClassPage explore_rows_frame={
    { { NULL },
        LABEL_MATRIX_BITS(1,1,0),
        LABEL_MATRIX_BITS(1,1,1),
        LABEL_MATRIX_BITS(1,1,2) },
    { LABEL_MATRIX_FREQ(1,1,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,0,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,0,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,0,2) },
    { LABEL_MATRIX_FREQ(1,1,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,1,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,1,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,1,2) },
    { LABEL_MATRIX_FREQ(1,1,2),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,2,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,2,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,2,2) },
    { LABEL_MATRIX_FREQ(1,1,3),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,3,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,3,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,3,2) },
    { LABEL_MATRIX_FREQ(1,1,4),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,4,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,4,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,4,2) },
    { LABEL_MATRIX_FREQ(1,1,5),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,5,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,5,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,5,2) },
    { LABEL_MATRIX_FREQ(1,1,6),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,6,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,6,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,6,2) },
    { LABEL_MATRIX_FREQ(1,1,7),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,7,0),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,7,1),
        CHECK_BUTTON_MATRIX(1,1,explore.enabled,bSupported,7,2) },
    { { NULL } }
  };

  static ClassPage explore_rows={
    { LABEL_DEVICE_WASAPI(1,0,explore.device) },
    { LABEL_STATIC(1,0,"Matrix of <i>WASAPI</i> Supported Integer Formats"
        " for Stereo (2.0)/Exclusive",NULL) },
    { FRAME(1,explore_rows_frame) },
    { { NULL } }
  };

  ////////////////
  static ClassNotebook notebook_pages={
    { LABEL_STATIC(1,0,"Common","Configure options independent of the"
        " chosen device."),NOTEBOOK_CONFIG(notebook_common_pages,notebook.common) },
    { LABEL_STATIC(1,0,"WASAPI","Configure options specific to the"
        " chosen device w.r.t <i>Microsoft WASAPI</i>."), PAGE(wasapi_rows) },
    { LABEL_STATIC(1,0,"YASAPI/NT","Configure options specific to the"
        " chosen device w.r.t the <i>YASAPI/NT</i> plug-in."),
            PAGE(yasapi_rows) },
    { LABEL_STATIC(1,0,"Monitor","Monitor running <i>YASAPI/NT</i>. Nothing"
        " to configure."),NOTEBOOK_CONFIG(notebook_monitor_pages,notebook.monitor) },
    { LABEL_STATIC(1,0,"Explore","Explore capabilities of <i>Microsoft</i>"
        " <i>WASAPI</i> w.r.t the chosen device. Nothing to configure."),
        PAGE(explore_rows) },
    { { NULL } }
  };

  ////////////////
  static ClassPage rows={
    { PAGE(general_rows) },
    { NOTEBOOK_CONFIG(notebook_pages,notebook.root) },
    { BUTTON_BOX() },
    { { NULL } }
  };

  static ClassCell cell=PAGE(rows);

  /////////////////////////////////////////////////////////////////////////////
  GtkWidget *window,*widget;
  Expand e;

  // window.
  pDialog->window = window = gtk_window_new (GTK_WINDOW_TOPLEVEL);

  // box.
  e.h = FALSE;
  e.v = FALSE;
DVWRITELN("%s {",cell.lpVtbl->pId);
  widget = cell.lpVtbl->create (&cell,pDialog,&e);
DVWRITELN("%s }",cell.lpVtbl->pId);
  g_object_set (G_OBJECT (widget), "margin", ePanelSpace, NULL);

  if (e.h) {
    gtk_widget_set_hexpand (widget, TRUE);
    gtk_widget_set_halign (widget, GTK_ALIGN_FILL);
  }
  else
    gtk_widget_set_halign (widget, GTK_ALIGN_CENTER);

  if (e.v) {
    gtk_widget_set_vexpand (widget, TRUE);
    gtk_widget_set_valign (widget, GTK_ALIGN_FILL);
  }
  else
    gtk_widget_set_valign (widget, GTK_ALIGN_CENTER);

  gtk_container_add (GTK_CONTAINER (window), widget);
  DWRITELN("\"box\" added to \"window\".");

  // Initially we need to apply the side effects.
  ConfigApply(pDialog,-1,"se");

  // It's better to call "set()" on all of our widgets after all of them
  // have been initialized because some of them may have side effects on
  // others.
  pDialog->lpVtbl->set(pDialog);
}

static void ConfigApply(Dialog *pDialog, int moffs, const char *op)
{
  enum {
    eWidgetSize=sizeof (ConfigWidget),
    eNumWidgetsGeneral=sizeof (pDialog->config.general)/eWidgetSize,
    eNumWidgetsCommon=sizeof (pDialog->config.common)/eWidgetSize,
    eNumWidgetsWasapi=sizeof (pDialog->config.wasapi)/eWidgetSize,
    eNumWidgetsYasapi=sizeof (pDialog->config.yasapi)/eWidgetSize,
    eNumWidgetsMonitor=sizeof (pDialog->config.monitor)/eWidgetSize,
    eNumWidgetsExplore=sizeof (pDialog->config.explore)/eWidgetSize,
    eNumWidgets
        =eNumWidgetsGeneral
        +eNumWidgetsCommon
        +eNumWidgetsWasapi
        +eNumWidgetsYasapi
        +eNumWidgetsMonitor
        +eNumWidgetsExplore,
  };

  typedef void (method_t)(ConfigWidget *pWidget);

  ConfigWidget *pWidget=pDialog->config.widgets;
  ConfigWidget *pWidgetMax=pWidget + eNumWidgets;
  method_t *method;

  while (pWidget<pWidgetMax) {
    if (moffs<0)
      method=pWidget->pCell->SideEffect;
    else if (pWidget->lpVtbl)
      method=*(method_t **)((char *)pWidget->lpVtbl+moffs);
    else
      method=NULL;

    if (method) {
      method(pWidget);
    }

    ++pWidget;
  }
}

static void ConfigSet(Dialog *pDialog)
{
  ConfigApply(pDialog,offsetof(IConfigWidgetVtbl,set),"set");
}

static void ConfigGet(Dialog *pDialog)
{
  ConfigApply(pDialog,offsetof(IConfigWidgetVtbl,get),"get");
}

static int ConfigPropertiesRead(Dialog *pDialog, Player *pPlayer)
{
  OptionsMemberConfig *pMember=&pDialog->config.options;
  const wchar_t *key=pDialog->lpVtbl->pIdW;

  PropertiesRead(gcaPropertiesConfig,pPlayer->ini,key,&pMember->original,
      &gcOptionsConfig);
  pMember->current=pMember->original;

  return 0;
}

static int ConfigPropertiesWrite(Dialog *pDialog, Player *pPlayer)
{
  OptionsMemberConfig *pMember=&pDialog->config.options;
  const wchar_t *key=pDialog->lpVtbl->pIdW;

  gtk_window_get_position (GTK_WINDOW (pDialog->window),
      &pMember->current.nPosX,
      &pMember->current.nPosY);
  pMember->current.nPage
      = gtk_notebook_get_current_page (GTK_NOTEBOOK (pDialog->config.notebook.root));
  pMember->current.common.nPage
      = gtk_notebook_get_current_page (GTK_NOTEBOOK (pDialog->config.notebook.common));
  pMember->current.monitor.nPage
      = gtk_notebook_get_current_page (GTK_NOTEBOOK (pDialog->config.notebook.monitor));
  PropertiesWrite(gcaPropertiesConfig,pPlayer->ini,key,
      &pMember->current,&pMember->original);

  return 0;
}

static void ConfigFinalize(Dialog *pDialog)
{
  int nPosX=pDialog->config.options.current.nPosX;
  int nPosY=pDialog->config.options.current.nPosY;
  int nPage;
  GtkWidget *notebook;

  if (0<=nPosX&&0<=nPosY)
    gtk_window_move (GTK_WINDOW(pDialog->window), nPosX, nPosY);

  nPage=pDialog->config.options.current.nPage;
  notebook=pDialog->config.notebook.root;

  if (0<=nPage)
    gtk_notebook_set_current_page (GTK_NOTEBOOK (notebook),nPage);

  nPage=pDialog->config.options.current.common.nPage;
  notebook=pDialog->config.notebook.common;

  if (0<=nPage)
    gtk_notebook_set_current_page (GTK_NOTEBOOK (notebook),nPage);

  nPage=pDialog->config.options.current.monitor.nPage;
  notebook=pDialog->config.notebook.monitor;

  if (0<=nPage)
    gtk_notebook_set_current_page (GTK_NOTEBOOK (notebook),nPage);
}

IDialogVtbl gIConfigVtbl={
  .pIdW=L"Config",
  .pTitle="Configure " YASAPI_TITLE_U8,
  .nWidthPixels=eConfigWidthPixels,
  .WindowCreate=ConfigWindowCreate,
  .set=ConfigSet,
  .get=ConfigGet,
  .PropertiesRead=ConfigPropertiesRead,
  .PropertiesWrite=ConfigPropertiesWrite,
  .Finalize=ConfigFinalize,
};
