/*
 * yasapi_config_combo_box.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>
#include <yasapi_config.h>

static void ConfigComboBoxSet(ConfigWidget *pWidget)
{
  ClassCell *pCell=pWidget->pCell;
  int *value=(int *)pCell->nav(pWidget->pDialog,pCell->offs);
  ConfigComboBoxMap *map;
  int i;

  for (map=pCell->data,i=0;map->label;++map,++i) {
    if (map->value==*value) {
      gtk_combo_box_set_active (GTK_COMBO_BOX (pWidget->widget), i);
      return;
    }
  }
}

static void ConfigComboBoxGet(ConfigWidget *pWidget)
{
  ClassCell *pCell=pWidget->pCell;
  int active = gtk_combo_box_get_active (GTK_COMBO_BOX (pWidget->widget));
  ConfigComboBoxMap *map = (ConfigComboBoxMap *)pCell->data + active;
  int *value;

  value=(int *)pCell->nav(pWidget->pDialog,pCell->offs);
  *value = map->value;

  if (pCell->SideEffect)
    pCell->SideEffect(pWidget);
}

static void yasapi_config_combo_box_changed (GtkComboBox *widget,
    gpointer user_data)
{
  Dialog *pDialog=user_data;
  ConfigWidget *pWidget=(ConfigWidget *)g_object_get_data(G_OBJECT(widget),
      CONFIG_WIDGET);

  pWidget->lpVtbl->get(pWidget);
  pDialog->lpVtbl->set(pDialog);
}

static GtkWidget *ConfigComboBoxCreate(ClassCell *pCell, Dialog *pDialog,
    Expand *pExpand)
{
  static IConfigWidgetVtbl vtbl={
    .set=ConfigComboBoxSet,
    .get=ConfigComboBoxGet,
  };

  GtkWidget *widget = gtk_combo_box_text_new();
  ConfigComboBoxMap *map;

  for (map=pCell->data;map->label;++map) {
    gtk_combo_box_text_append (GTK_COMBO_BOX_TEXT (widget), NULL, 
        map->label);
  }

  ConfigWidgetCreate(pCell, pDialog, &vtbl, widget);

  g_signal_connect (widget, "changed",
      G_CALLBACK (yasapi_config_combo_box_changed), pDialog);

  return widget;
}

IClassCellVtbl gcIClassComboBoxVtbl={
  .create=ConfigComboBoxCreate,
};
