/*
 * yasapi_config_progess_bar.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>
#include <yasapi_config.h>

static void ConfigProgessBarSet(ConfigWidget *pWidget)
{
  ClassCell *pCell=pWidget->pCell;
  double *pxValue=(double *)pCell->nav(pWidget->pDialog,pCell->offs);

  if (pxValue&&!YASAPI_DOUBLE_INVALID(*pxValue)) {
    gtk_progress_bar_set_fraction  (GTK_PROGRESS_BAR (pWidget->widget),
        *pxValue);
  }
  else
    gtk_progress_bar_set_fraction  (GTK_PROGRESS_BAR (pWidget->widget), 0.0);
}

static void ConfigProgessBarGet(ConfigWidget *pWidget)
{
}

static GtkWidget *ConfigProgessBarCreate(ClassCell *pCell, Dialog *pDialog,
    Expand *pExpand)
{
  static IConfigWidgetVtbl vtbl={
    .set=ConfigProgessBarSet,
    .get=ConfigProgessBarGet,
  };

  GtkWidget *widget;

  widget = gtk_progress_bar_new ();

  ConfigWidgetCreate(pCell, pDialog, &vtbl, widget);
  pExpand->h=TRUE;

  return widget;
}

IClassCellVtbl gcIClassProgessBarVtbl={
  .create=ConfigProgessBarCreate,
};

static void ConfigProgessBarSetStyle(ConfigWidget *pWidget,
    ConfigColorTheme *pTheme)
{
  // For GTK 3.20 and newer, when defining the CSS, it seems that, instead
  // of using the progressbar class, you must use the progress and trough
  // classes. This is what worked for me:
  //
  // progress, trough {
  //    min-height: 30px;
  // }
  //
  // Gtk Progress bar height
  // https://stackoverflow.com/questions/44290651/gtk-progress-bar-height
  static char format_height_color[]=
    "progress, trough {"
    "\n  min-height: %dpx;"
    "\n}"
    "\nprogress {"
    "\n  background-color: rgb(%d,%d,%d);"
    "\n  border-color: rgb(%d,%d,%d);"
    "\n}"
    ;
  static char format_height[]=
    "progress, trough {"
    "\n  min-height: %dpx;"
    "\n}"
    ;
  ConfigProgessBarParameter *p=pWidget->pCell->data;
  GtkStyleContext *context = gtk_widget_get_style_context(pWidget->widget);
  char data[256];

  if (pWidget->provider) {
    gtk_style_context_remove_provider (context,
        GTK_STYLE_PROVIDER(pWidget->provider));
    pWidget->provider=NULL;
  }

  if (pTheme) {
    sprintf(data,format_height_color,
        p->height?p->height:16,
        pTheme->fill.red,pTheme->fill.green,pTheme->fill.blue,
        pTheme->border.red,pTheme->border.green,pTheme->border.blue);
  }
  else
    sprintf(data,format_height,p->height?p->height:16);

  pWidget->provider = gtk_css_provider_new();
  gtk_css_provider_load_from_data (pWidget->provider, data, -1, NULL);
  gtk_style_context_add_provider (context,
      GTK_STYLE_PROVIDER (pWidget->provider),
      GTK_STYLE_PROVIDER_PRIORITY_USER);
}

///////////////////////////////////////////////////////////////////////////////
static ConfigColorTheme red_light={
  .fill={ .red=255,.green=128,.blue=0 },
  .border={ .red=204,.green=102,.blue=0 }
};

static ConfigColorTheme red={
  .fill={ .red=255,.green=0,.blue=0 },
  .border={ .red=196,.green=0,.blue=0 }
};

static ConfigColorTheme green_light={
  .fill={ .red=181,.green=230,.blue=29 },
  .border={ .red=136,.green=174,.blue=19 }
};

static ConfigColorTheme green={
  .fill={ .red=34,.green=177,.blue=76 },
  .border={ .red=29,.green=150,.blue=65 }
};

static ConfigColorTheme blue_light={
  .fill={ .red=153,.green=217,.blue=234 },
  .border={ .red=65,.green=186,.blue=216 }
};

void ConfigMonitorEngineClientTailSetStyle(ConfigWidget *pWidget)
{
  Dialog *pDialog=pWidget->pDialog;
  ConfigWidget *pWidgetMin=&pDialog->config.monitor.engine.tail_min;
  ConfigWidget *pWidgetCur=&pDialog->config.monitor.engine.tail_cur;
  ConfigWidget *pWidgetMax=&pDialog->config.monitor.engine.tail_max;

  switch (DialogOptionsCommonConfig(pDialog)->eThemeEngineTail) {
  case eConfigThemeRed:
    ////////
    ConfigProgessBarSetStyle(pWidgetMin,&red_light);
    ConfigProgessBarSetStyle(pWidgetCur,&red);
    ConfigProgessBarSetStyle(pWidgetMax,&red_light);
    break;
  case eConfigThemeGreen:
    ConfigProgessBarSetStyle(pWidgetMin,&green_light);
    ConfigProgessBarSetStyle(pWidgetCur,&green);
    ConfigProgessBarSetStyle(pWidgetMax,&green_light);
    break;
  case eConfigThemeBlue:
    ConfigProgessBarSetStyle(pWidgetMin,&blue_light);
    ConfigProgessBarSetStyle(pWidgetCur,NULL);
    ConfigProgessBarSetStyle(pWidgetMax,&blue_light);
    break;
  default:
    ConfigProgessBarSetStyle(pWidgetMin,NULL);
    ConfigProgessBarSetStyle(pWidgetCur,NULL);
    ConfigProgessBarSetStyle(pWidgetMax,NULL);
    break;
  }
}

void ConfigMonitorPaddingSetStyle(ConfigWidget *pWidget)
{
  Dialog *pDialog=pWidget->pDialog;

  switch (DialogOptionsCommonConfig(pDialog)->eThemePadding) {
  case eConfigThemeRed:
    ConfigProgessBarSetStyle(&pDialog->config.monitor.padding,&red);
    break;
  case eConfigThemeGreen:
    ConfigProgessBarSetStyle(&pDialog->config.monitor.padding,&green);
    break;
  default:
    ConfigProgessBarSetStyle(&pDialog->config.monitor.padding,NULL);
    break;
  }
}

void ConfigMonitorStartSetStyle(ConfigWidget *pWidget)
{
  Dialog *pDialog=pWidget->pDialog;

  switch (DialogOptionsCommonConfig(pDialog)->eThemeStart) {
  case eConfigThemeRed:
    ConfigProgessBarSetStyle(&pDialog->config.monitor.start,&red);
    break;
  case eConfigThemeGreen:
    ConfigProgessBarSetStyle(&pDialog->config.monitor.start,&green);
    break;
  default:
    ConfigProgessBarSetStyle(&pDialog->config.monitor.start,NULL);
    break;
  }
}

void ConfigMonitorWrittenSetStyle(ConfigWidget *pWidget)
{
  Dialog *pDialog=pWidget->pDialog;

  switch (DialogOptionsCommonConfig(pDialog)->eThemeWritten) {
  case eConfigThemeRed:
    ConfigProgessBarSetStyle(&pDialog->config.monitor.written,&red);
    break;
  case eConfigThemeGreen:
    ConfigProgessBarSetStyle(&pDialog->config.monitor.written,&green);
    break;
  default:
    ConfigProgessBarSetStyle(&pDialog->config.monitor.written,NULL);
    break;
  }
}
