/*
 * yasapi_producer_params.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
#if ! defined (AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM) // [
#define AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM 0x80000000
#endif // ]

#if ! defined (AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY) // [
#define AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY 0x08000000
#endif // ]

///////////////////////////////////////////////////////////////////////////////
int EngineParamCreate(EngineParam *pParam, const EngineOptions *pOptions,
    const EngineClient *pClient)
{
  int code=-1;
  Cache *pCache=&pClient->pPlayer->cache;

  /////////////////////////////////////////////////////////////////////////////
  memset(pParam,0,sizeof *pParam);
  pParam->pDevice=pCache->selected.active.pCacheDevice->pDevice;
  pParam->RingCreate=RingCreateFrames;
  pParam->bActive=1;

  if (pOptions->common.bLatency)
    pParam->time.lpVtbl=&gcIEngineServerTimeNullVtbl;
  else
    pParam->time.lpVtbl=&gcIEngineServerTimeClockVtbl;


  /////////////////////////////////////////////////////////////////////////////
  pParam->pFriendlyName
      =pCache->selected.active.pCacheDevice->pFriendlyNameOriginal;

  if (pOptions->device.eStrategy) {
    pParam->dwStreamFlags=AUDCLNT_STREAMFLAGS_EVENTCALLBACK;
    DWRITELN("AUDCLNT_STREAMFLAGS_EVENTCALLBACK");
  }
  else
    pParam->dwStreamFlags=0ul;

  if (pOptions->device.eMode) {
    pParam->eShareMode=AUDCLNT_SHAREMODE_EXCLUSIVE;
    DWRITELN("AUDCLNT_SHAREMODE_EXCLUSIVE");
  }
  else {
    pParam->eShareMode=AUDCLNT_SHAREMODE_SHARED;
    DWRITELN("AUDCLNT_SHAREMODE_SHARED");

    if (!pOptions->device.bAutoConvertPCM)
      goto shared;

    pParam->dwStreamFlags|=AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM;
    DWRITELN("AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM");

    if (!pOptions->device.bSRCDefaultQuality)
      goto shared;

    pParam->dwStreamFlags|=AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY;
    DWRITELN("AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY");
  shared: ;
  }

  DVWRITELN("dwStreamFlags=%0lx",pParam->dwStreamFlags);
  pParam->bDoubleBuffered=pOptions->device.eMode&&pOptions->device.eStrategy;

  /////////////////////////////////////////////////////////////////////////////
  switch (pOptions->common.eVolume) {
  case eVolumeSoftware:
    pParam->volume.lpVtbl=&gcIEngineServerVolumeSoftwareVtbl;
    break;
  case eVolumeWasapi:
    switch (pOptions->device.eMode) {
    case eModeShared:
      pParam->volume.lpVtbl=&gcIEngineServerVolumeSimpleAudioVtbl;
      break;
    case eModeExclusive:
      pParam->volume.lpVtbl=&gcIEngineServerVolumeAudioEndpointVtbl;
      break;
    default:
      pParam->volume.lpVtbl=&gcIEngineServerVolumeSoftwareVtbl;
      break;
    }

    break;
  default:
    pParam->volume.lpVtbl=&gcIEngineServerVolumeNullVtbl;
    break;
  };

  code=0;
  return code;
}
