/*
 * yasapi_gtk_client.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>

int GtkClientCreate(GtkClient *pClient, Player *pPlayer)
{
  FsmCreate(pClient,gcaIGtkClientState ,NULL,NULL);
  pClient->pPlayer=pPlayer;

  return 0;
}

void GtkClientDestroy(GtkClient *pClient)
{
  YASAPI_PLAYER_LOCK(pClient->pPlayer); // {
  FsmDestroy(pClient);
  YASAPI_PLAYER_UNLOCK(pClient->pPlayer); // }
}

int GtkClientShow(GtkClient *pClient, Dialog *pDialog, HWND hWndParent)
{
  int code=-1;
  Player *pPlayer=pClient->pPlayer;

  YASAPI_PLAYER_LOCK(pPlayer); // {
  pDialog->hWndParent=hWndParent;
  pClient->pDialog=pDialog;
  
  if (FsmState(pClient)<eGtkClientStateThreadRunning) {
    if (YASAPI_FSM_UP(pClient,eGtkClientStateThreadRunning)<0) {
      DMESSAGE("starting gtk");
      goto exit;
    }
  }
  else if (pDialog->window)
    g_idle_add (GtkServerDialogShowAsync, pDialog);

  code=0;
exit:
  YASAPI_PLAYER_UNLOCK(pPlayer); // }
  return code;
}

static gboolean GtkClientWidgetUpdateAsync(gpointer user_data)
{
  ConfigWidget *pWidget=user_data;
  Player *pPlayer=pWidget->pDialog->pGtk->pPlayer;

  YASAPI_PLAYER_LOCK(pPlayer); // {
  pWidget->lpVtbl->set(pWidget);
  YASAPI_PLAYER_UNLOCK(pPlayer); // }

  return FALSE;
}

void GtkClientUpdateWidget(GtkClient *pClient, ConfigWidget *pWidget)
{
  GtkServer *pServer;

  if (!pWidget->lpVtbl)
    return;

  if (!(pServer=pClient->pServer))
    return;

  if (FsmState(pServer)<eGtkServerStateReady)
    return;

  g_idle_add (GtkClientWidgetUpdateAsync, pWidget);
}

static gboolean DialogUpdateAsync(gpointer user_data)
{
  Dialog *pDialog=user_data;
  Player *pPlayer=pDialog->pGtk->pPlayer;

  YASAPI_PLAYER_LOCK(pPlayer); // {
  pDialog->lpVtbl->set(pDialog);
  YASAPI_PLAYER_UNLOCK(pPlayer); // }

  return FALSE;
}

void GtkClientUpdateDialog(GtkClient *pClient, Dialog *pDialog)
{
  GtkServer *pServer;

  if (!pDialog->lpVtbl)
    return;

  if (!(pServer=pClient->pServer))
    return;

  if (FsmState(pServer)<eGtkServerStateReady)
    return;

  g_idle_add (DialogUpdateAsync, pDialog);
}
