/*
 * yasapi_gtk_client_state.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
static int GtkClientStateCacheEnter(GtkClient *pClient)
{
  if (YASAPI_FSM_UP(pClient->pPlayer,ePlayerStateCache)<0)
    goto error;

  ++pClient->fsm.pState;

  return 0;
error:
  return -1;
}

static void GtkClientStateCacheLeave(GtkClient *pClient)
{
  --pClient->fsm.pState;
}

///////////////////////////////////////////////////////////////////////////////
// We subclass the parent window in order to display a wait cursor.
static LRESULT CALLBACK GtkWndProc(HWND hWnd, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
  switch (uMsg) {
  case WM_SETCURSOR:
    SetCursor(LoadCursor(0,IDC_WAIT));
    return TRUE;
  default:
    break;
  }

  return DefWindowProc(hWnd,uMsg,wParam,lParam);
}

static int GtkClientStateSubclassEnter(GtkClient *pClient)
{
  if (!pClient->pDialog->hWndParent) {
    DMESSAGE("subclassing");
    goto exit;
  }

  pClient->hWndParent=pClient->pDialog->hWndParent;

  pClient->lpPrevWndFunc=(WNDPROC)SetWindowLong(
    pClient->hWndParent,            // _In_ HWND hWnd,
    GWL_WNDPROC,                    // _In_ int  nIndex,
    (LONG)GtkWndProc                // _In_ LONG dwNewLong
  );

  if (!pClient->lpPrevWndFunc) {
    DMESSAGE("subclassing");
    goto exit;
  }

  ++pClient->fsm.pState;

  return 0;
exit:
  return -1;
}

static void GtkClientStateSubclassLeave(GtkClient *pClient)
{
  --pClient->fsm.pState;
}

///////////////////////////////////////////////////////////////////////////////
static int GtkClientStateThreadEventEnter(GtkClient *pClient)
{
  // This is a MANUAL RESET event!
  if (!(pClient->thread.hEvent=CreateEventW(NULL,TRUE,FALSE,NULL))) {
    DMESSAGE("creating consumer event");
    goto error;
  }

  ++pClient->fsm.pState;

  return 0;
error:
  return -1;
}

static void GtkClientStateThreadEventLeave(GtkClient *pClient)
{
  --pClient->fsm.pState;
  CloseHandle(pClient->thread.hEvent);
}

///////////////////////////////////////////////////////////////////////////////
static int GtkClientStateThreadEnter(GtkClient *pClient)
{
  pClient->thread.hThread=CreateThread(
    NULL,       // _In_opt_  LPSECURITY_ATTRIBUTES  lpThreadAttributes,
    0ul,        // _In_      SIZE_T                 dwStackSize,
    GtkServerThreadProc,
                // _In_      LPTHREAD_START_ROUTINE lpStartAddress,
    pClient,  // _In_opt_  LPVOID                 lpParameter,
    0ul,        // _In_      DWORD                  dwCreationFlags,
                // _Out_opt_ LPDWORD                lpThreadId
    &pClient->thread.dwThreadId
  );

  if (INVALID_HANDLE_VALUE==pClient->thread.hThread) {
    DMESSAGE("starting consumer thread");
    goto error;
  }

  ++pClient->fsm.pState;

  return 0;
error:
  return -1;
}

static void GtkClientStateThreadLeave(GtkClient *pClient)
{
  --pClient->fsm.pState;
  WaitForSingleObject(pClient->thread.hThread,INFINITE);
  CloseHandle(pClient->thread.hThread);
  pClient->thread.hThread=INVALID_HANDLE_VALUE;
}

///////////////////////////////////////////////////////////////////////////////
static int GtkClientStateThreadRunningEnter(GtkClient *pClient)
{
  IFsmState *pState=pClient->fsm.pStateMin+eGtkClientStateThreadRunning;

  while (pClient->fsm.pState<pState) {
    ResetEvent(pClient->thread.hEvent);
    WaitForSingleObject(pClient->thread.hEvent,INFINITE);
  }
  DWRITELN("!!! GTK THREAD RUNNING !!!!!!!!!!!!!!!!!!!!!!");

  return 0;
}

static void GtkClientStateThreadRunningLeave(GtkClient *pClient)
{
  IFsmState *pState=pClient->fsm.pStateMin+eGtkClientStateThreadRunning-1;
  Dialog *pDialogMax=pClient->aDialog+eGtkNumDialogs;
  Dialog *pDialog;

  for (pDialog=pClient->aDialog;pDialog<pDialogMax;++pDialog)
    GtkServerDialogWrite(pDialog);

  gtk_main_quit ();

  while (pState<pClient->fsm.pState) {
    ResetEvent(pClient->thread.hEvent);
    WaitForSingleObject(pClient->thread.hEvent,INFINITE);
  }
}

///////////////////////////////////////////////////////////////////////////////
IGtkClientState gcaIGtkClientState[eGtkClientStateCount]={
#define YASAPI_GTK_CLIENT(x) { \
  .pId=#x, \
  .Enter=GtkClientState ## x ## Enter, \
  .Leave=GtkClientState ## x ## Leave, \
},
#include <yasapi_gtk_client_state.h>
#undef YASAPI_GTK_CLIENT
};
