/*
 * yasapi_gtk_server_state.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>

#if defined (YASAPI_PLUGIN_DIRECTORY) // [
#include <winamp/wa_ipc.h>

#define IPC_GETPLUGINDIRECTORYW 1336
#endif // ]

///////////////////////////////////////////////////////////////////////////////
static int GtkServerStateThreadEventEnter(GtkServer *pServer)
{
  // This is a MANUAL RESET event!
  if (!(pServer->thread.hEvent=CreateEventW(NULL,TRUE,FALSE,NULL))) {
    DMESSAGE("creating consumer event");
    goto exit;
  }

  ++pServer->fsm.pState;

  return 0;
exit:
  return -1;
}

static void GtkServerStateThreadEventLeave(GtkServer *pServer)
{
  --pServer->fsm.pState;
  CloseHandle(pServer->thread.hEvent);
}

///////////////////////////////////////////////////////////////////////////////
static int GtkServerStateOptionsCommonEnter(GtkServer *pServer)
{
  GtkClient *pClient=pServer->pClient;

DVWRITELNW(L"\"%s\"",pClient->pPlayer->ini);
  PropertiesRead(gcaPropertiesGtk,pClient->pPlayer->ini,
      gcaKeyGtk,&pServer->options.original,&gcOptionsGtk);
  pServer->options.current=pServer->options.original;
DVWRITELNW(L"\"%s\"",pServer->options.current.path);
  ++pServer->fsm.pState;

  return 0;
}

static void GtkServerStateOptionsCommonLeave(GtkServer *pServer)
{
  --pServer->fsm.pState;

DVWRITELNW(L"\"%s\"",pServer->options.current.path);
  PropertiesWrite(gcaPropertiesGtk,pServer->pClient->pPlayer->ini,
     gcaKeyGtk,&pServer->options.current,&pServer->options.original);
}

///////////////////////////////////////////////////////////////////////////////
typedef struct _GtkLib GtkLib;

struct _GtkLib {
  const char *name;
  const gchar *(*check) (guint required_major, guint required_minor,
      guint required_micro);
  guint major;
  guint minor;
  guint micro;
  int bLibGlib;
};

static GtkLib gaGtkLib[eGtkLibCount]={
  { "libglib-2.0-0.dll",glib_check_version,
      GLIB_MAJOR_VERSION,GLIB_MINOR_VERSION,GLIB_MICRO_VERSION,1 },
  { "libgobject-2.0-0.dll",NULL,0,0,0 },
  { "libgdk-3-0.dll",NULL,
      GDK_MAJOR_VERSION,GDK_MINOR_VERSION,GDK_MICRO_VERSION },
  { "libgtk-3-0.dll",gtk_check_version,
      GTK_MAJOR_VERSION,GTK_MINOR_VERSION,GTK_MICRO_VERSION },
};

static void GtkServerUnloadLibs(GtkServer *pServer)
{
  while (pServer->ahGtkLib<pServer->phGtkLib) {
    FreeLibrary(*--pServer->phGtkLib);
    DVWRITELN("\"%s\" unloaded.",gaGtkLib[pServer->phGtkLib-pServer->ahGtkLib].name);
  }
}

static int GtkServerShowPathDialog(GtkServer *pServer, const char *msg)
{
	enum { eTitleNumChars=256 };
  static const wchar_t TITLE[]=
#if defined (YASAPI_FILE_OPEN_DIALOG) // [
			L" "
#else // ] [
			L"\n"
#endif // ]
			L"Choose the GTK Runtime installation's root"
#if defined (YASAPI_FILE_OPEN_DIALOG) // [
			L" "
#else // ] [
			L"\n"
#endif // ]
			L"(i.e. the directory on top of \"bin\"):";

  int code=-1;
  wchar_t *path=pServer->options.current.path;
  wchar_t title[eTitleNumChars];
  int nSize;
#if defined (YASAPI_FILE_OPEN_DIALOG) // [
  HWND hWndParent=pServer->pClient->pDialog->hWndParent;
  IFileDialog *pFileDialog;
  IShellItem *pItem;
  PWSTR pFilePath;
  HRESULT hr;
#endif // ]

  nSize=MultiByteToWideChar(
    CP_UTF8,      		// UINT                              CodePage,
    0ul,          		// DWORD                             dwFlags,
    msg,          		// _In_NLS_string_(cbMultiByte)LPCCH lpMultiByteStr,
    -1,           		// int                               cbMultiByte,
    title,        		// LPWSTR                            lpWideCharStr,
    eTitleNumChars    // int                               cchWideChar
  );

	--nSize;
  wcscpy_s(title+nSize,eTitleNumChars-nSize,TITLE);

#if defined (YASAPI_FILE_OPEN_DIALOG) // [
  /////////////////////////////////////////////////////////////////////////////
  GtkServerUnloadLibs(pServer);
  *path=0;

  /////////////////////////////////////////////////////////////////////////////
  hr=CoCreateInstance(
    &CLSID_FileOpenDialog,    // _In_  REFCLSID  rclsid,
    NULL,                     // _In_  LPUNKNOWN pUnkOuter,
    CLSCTX_INPROC_SERVER,     // _In_  DWORD     dwClsContext,
    &IID_IFileDialog,         // _In_  REFIID    riid,
    (void **)&pFileDialog     // _Out_ LPVOID    *ppv
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("CoCreateInstance",REGDB_E_CLASSNOTREG);
    DCMESSAGE("CoCreateInstance",CLASS_E_NOAGGREGATION);
    DCMESSAGE("CoCreateInstance",E_NOINTERFACE);
    DCMESSAGE("CoCreateInstance",E_POINTER);
    DDMESSAGE("CoCreateInstance",hr);
    }

    goto ecreate;
  }

  /////////////////////////////////////////////////////////////////////////////
  hr=pFileDialog->lpVtbl->SetTitle(pFileDialog,title);

  if (FAILED(hr))
    goto etitle;

  /////////////////////////////////////////////////////////////////////////////
  hr=pFileDialog->lpVtbl->SetOptions(pFileDialog,FOS_PICKFOLDERS);

  if (FAILED(hr))
    goto eoptions;

  /////////////////////////////////////////////////////////////////////////////
  hr=pFileDialog->lpVtbl->Show(pFileDialog,hWndParent);

  if (FAILED(hr))
    goto eshow;

  /////////////////////////////////////////////////////////////////////////////
  hr=pFileDialog->lpVtbl->GetResult(pFileDialog,&pItem);

  if (FAILED(hr))
    goto eitem;

  /////////////////////////////////////////////////////////////////////////////
  hr=pItem->lpVtbl->GetDisplayName(pItem,SIGDN_FILESYSPATH,&pFilePath);

  if (FAILED(hr))
    goto epath;

  wcscpy_s(path,YASAPI_ARRAYSIZE(pServer->options.current.path),pFilePath);

  /////////////////////////////////////////////////////////////////////////////
  code=0;
  CoTaskMemFree(pFilePath);
epath:
  pItem->lpVtbl->Release(pItem);
eitem:
eshow:
eoptions:
etitle:
  pFileDialog->lpVtbl->Release(pFileDialog);
ecreate:
#else // ] [
  // https://www.codeproject.com/Articles/13088/How-to-Browse-for-a-Folder
  BROWSEINFOW bi={ 0 };
  LPITEMIDLIST pIdl;

  bi.ulFlags=BIF_USENEWUI;
  bi.lpszTitle = title;
  pIdl=SHBrowseForFolderW(&bi);

  if (!pIdl)
    goto exit;

  SHGetPathFromIDListW(pIdl,path);
  CoTaskMemFree(pIdl);
  code=0;
exit:
#endif // ]
  return code;
}

static int GtkServerStateLoadEnter(GtkServer *pServer)
{
#if defined (YASAPI_PLUGIN_DIRECTORY) // [
  static const wchar_t aPathName[]=L"\\yagtk-rt\\bin";
#else // ] [
  static const wchar_t aPathName[]=L"yagtk-rt\\bin";
#endif // ]

  GtkClient *pClient=pServer->pClient;
  const GtkLib *pClientLib=gaGtkLib;
  const wchar_t *pPath;
  wchar_t path[MAX_PATH],*pp;
#if defined (YASAPI_PLUGIN_DIRECTORY) // [
  wchar_t *plugin;
#else // ] [
  DWORD dwSize;
#endif // ]
  char msg[256];
  const gchar *check;
retry:
  pPath=pServer->options.current.path;
  pServer->phGtkLib=pServer->ahGtkLib;

  if (*pPath) {
    ///////////////////////////////////////////////////////////////////////////
    pp=path;
    wcscpy(pp,pPath);
    pp+=wcslen(pp);
    wcscpy(pp,L"\\bin");
  }
  else if (pClient->pPlayer->pOut) {
    ///////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_PLUGIN_DIRECTORY) // [
    plugin=(wchar_t *)SendMessageW(pClient->pPlayer->pOut->hMainWindow,
        WM_WA_IPC,0,IPC_GETPLUGINDIRECTORYW);

    wcsncpy(
      path,                    // wchar_t *strDest,
      plugin,
                            // const wchar_t *strSource,
      (sizeof path)/(sizeof path[0])
                            // size_t count
    );

    pp=path+wcslen(path);
#else // ] [
    dwSize=GetModuleFileNameW(
      pClient->pPlayer->pOut->hDllInstance,
                      // _In_opt_ HMODULE hModule,
      path,           // _Out_    LPTSTR  lpFilename,
      sizeof path     // _In_     DWORD   nSize
    );

    for (pp=path+dwSize;path<pp&&'\\'!=pp[-1];--pp)
      ;
#endif // ]

    wcscpy(pp,aPathName);
  }
  else {
    DMESSAGE("Missing path to library");
    goto exit;
  }

  if (!SetDllDirectoryW(path)) {
    DVMESSAGE("extending library path with error %lu",GetLastError());
    goto exit;
  }

  DVWRITELNW(L"Library search path extended by \"%s\".",path);

  /////////////////////////////////////////////////////////////////////////////
  while (pServer->phGtkLib<pServer->ahGtkLib+eGtkLibCount) {
    if (!(*pServer->phGtkLib=LoadLibraryA(pClientLib->name))) {
      sprintf(msg,"Error loading \"%s\".",pClientLib->name);

DMARKLN();
      if (GtkServerShowPathDialog(pServer,msg)<0)
        goto exit;

      goto retry;
    }

    DVWRITELN("\"%s\" loaded.",pClientLib->name);

    if (pClientLib->bLibGlib)
      pClient->hLibGlib=*pServer->phGtkLib;

    if (pClientLib->check)
      check=pClientLib->check(pClientLib->major,pClientLib->minor,pClientLib->micro);
    else
      check=NULL;

    ++pServer->phGtkLib;
    ++pClientLib;

    if (check) {
      sprintf(msg,"Error \"%s\" check failed: \"%s\"",pClientLib[-1].name,check);

DMARKLN();
      if (GtkServerShowPathDialog(pServer,msg)<0)
        goto exit;

      goto retry;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  gtk_init(NULL,NULL);
  DWRITELN("GTK initialized.");

  /////////////////////////////////////////////////////////////////////////////
  ++pServer->fsm.pState;

  return 0;
exit:
  return -1;
}

static void GtkServerStateLoadLeave(GtkServer *pServer)
{
  --pServer->fsm.pState;
  GtkServerUnloadLibs(pServer);
}

///////////////////////////////////////////////////////////////////////////////
static int GtkServerStatePropertiesDialogsEnter(GtkServer *pServer)
{
  GtkClient *pClient=pServer->pClient;
  Dialog *pDialogMax=pClient->aDialog+eGtkNumDialogs;
  Dialog *pDialog;

  pClient->about.lpVtbl=&gIAboutVtbl;
  pClient->about.pGtk=pClient;

  pClient->config.lpVtbl=&gIConfigVtbl;
  pClient->config.pGtk=pClient;

  for (pDialog=pClient->aDialog;pDialog<pDialogMax;++pDialog) {
    DASSERT(pDialog->lpVtbl,exit);
    DASSERT(pDialog->pGtk,exit);
    pDialog->bMoved=0;
    pDialog->lpVtbl->PropertiesRead(pDialog,pClient->pPlayer);
  }

  ++pServer->fsm.pState;

  return 0;
DLABEL(exit);
  return -1;
}

static void GtkServerStatePropertiesDialogsLeave(GtkServer *pServer)
{
  --pServer->fsm.pState;
}

///////////////////////////////////////////////////////////////////////////////
static int GtkServerStateDialogsEnter(GtkServer *pServer)
{
  GtkClient *pClient;
  Dialog *pDialogMax;
  Dialog *pDialog;

  pClient=pServer->pClient;
  pDialogMax=pClient->aDialog+eGtkNumDialogs;

  for (pDialog=pClient->aDialog;pDialog<pDialogMax;++pDialog) {
    pDialog->lpVtbl->WindowCreate (pDialog);
    gtk_window_set_title (GTK_WINDOW (pDialog->window),
        pDialog->lpVtbl->pTitle);
    gtk_window_set_default_size (GTK_WINDOW (pDialog->window),
        pDialog->lpVtbl->nWidthPixels, -1);
    gtk_window_set_position (GTK_WINDOW(pDialog->window),
        GTK_WIN_POS_CENTER);
    g_signal_connect (G_OBJECT(pDialog->window), "delete-event",
        G_CALLBACK(window_delete_event), pDialog);
  }

DMARKLN();
  if (pClient->pDialog)
    gtk_widget_show_all (pClient->pDialog->window);
DMARKLN();

  ++pServer->fsm.pState;

  return 0;
}

static void GtkServerStateDialogsLeave(GtkServer *pServer)
{
  --pServer->fsm.pState;
}

///////////////////////////////////////////////////////////////////////////////
static int GtkServerStatePresentEnter(GtkServer *pServer)
{
  Player *pPlayer=pServer->pClient->pPlayer;

  if (YASAPI_FSM_UP(&pPlayer->present,ePresentClientStateRunning)<0) {
    DMESSAGE("starting present client");
    goto exit;
  }

  ++pServer->fsm.pState;

  return 0;
exit:
  return -1;
}

static void GtkServerStatePresentLeave(GtkServer *pServer)
{
  Player *pPlayer=pServer->pClient->pPlayer;

  --pServer->fsm.pState;
  YASAPI_FSM_DOWN(&pPlayer->present,0);
}

///////////////////////////////////////////////////////////////////////////////
static int GtkServerStateUnsubclassEnter(GtkServer *pServer)
{
  GtkClient *pClient=pServer->pClient;

  SetWindowLong(
    pClient->hWndParent,          // _In_ HWND hWnd,
    GWL_WNDPROC,                  // _In_ int  nIndex,
    (LONG)pClient->lpPrevWndFunc  // _In_ LONG dwNewLong
  );

  pClient->hWndParent=NULL;
  pClient->lpPrevWndFunc=NULL;
  ++pServer->fsm.pState;

  return 0;
}

static void GtkServerStateUnsubclassLeave(GtkServer *pServer)
{
  --pServer->fsm.pState;
}

///////////////////////////////////////////////////////////////////////////////
IGtkServerState gcaIGtkServerState[eGtkServerStateCount]={
#define YASAPI_GTK_SERVER(x) { \
  .pId=#x, \
  .Enter=GtkServerState ## x ## Enter, \
  .Leave=GtkServerState ## x ## Leave, \
},
#include <yasapi_gtk_server_state.h>
#undef YASAPI_GTK_SERVER
};
