/*
 * yasapi_param.h
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#ifndef __YASAPI_PARAM_H__ // [
#define __YASAPI_PARAM_H__

#include <yasapi_trace.h>
#include <windows.h>
#include <mmdeviceapi.h>
#include <audiosessiontypes.h>
#include <audioclient.h>
#include <functiondiscoverykeys_devpkey.h>
#include <endpointvolume.h>
#if defined (YASAPI_FILE_OPEN_DIALOG) // [
#include <shobjidl.h>
#else // ] [
#include <shlobj.h>
#endif // ]
#include <gtk/gtk.h>
#include <gdk/gdkwin32.h>

#ifdef __cplusplus // [
extern "C" {
#endif // ]

///////////////////////////////////////////////////////////////////////////////
extern const CLSID CLSID_MMDeviceEnumerator;
extern const IID IID_IMMDeviceEnumerator;
extern const IID IID_IAudioClient;
extern const IID IID_IAudioClock;
extern const IID IID_IAudioRenderClient;
extern const IID IID_ISimpleAudioVolume;
extern const IID IID_IAudioEndpointVolume;
#if defined (YASAPI_FILE_OPEN_DIALOG) // [
extern const CLSID CLSID_FileOpenDialog;
extern const IID IID_IFileDialog;
#endif // ]

///////////////////////////////////////////////////////////////////////////////
typedef struct _Player Player;
typedef struct _Dialog Dialog;

typedef struct _OptionsGtk OptionsGtk;
typedef struct _OptionsAbout OptionsAbout;
typedef struct _OptionsConfig OptionsConfig;
typedef enum _EMono2Stereo EMono2Stereo;
typedef enum _EVolume EVolume;
typedef enum _EDither EDither;
typedef enum _EConfigTheme EConfigTheme;
typedef enum _EThreadPriority EThreadPriority;
typedef struct _OptionsCommon OptionsCommon;
typedef enum _EPeriod EPeriod;
#define YASAPI_MODE
#if defined (YASAPI_MODE) // [
typedef enum _EMode EMode;
#endif // ]
#define YASAPI_STRATEGY
#if defined (YASAPI_STRATEGY) // [
typedef enum _EStrategy EStrategy;
#endif // ]
typedef enum _EBitdepth EBitdepth;
typedef struct _OptionsDevice OptionsDevice;

typedef struct _OptionsMemberAbout OptionsMemberAbout;
typedef struct _OptionsMemberConfig OptionsMemberConfig;
typedef struct _OptionsMemberGtk OptionsMemberGtk;
typedef struct _OptionsMemberCommon OptionsMemberCommon;
typedef struct _OptionsMemberDevice OptionsMemberDevice;
typedef struct _Options Options;

typedef const struct _ICacheDeviceVtbl ICacheDeviceVtbl;
typedef struct _CacheDevice CacheDevice;
typedef struct _Cache Cache;

typedef const struct _IPropertyVtbl IPropertyVtbl;
typedef struct _Property Property;

typedef enum _EFormatMatrixLimits EFormatMatrixLimits;
typedef const struct _Frequency Frequency;
typedef const struct _BitDepth BitDepth;
typedef struct _FormatMatrix FormatMatrix;
typedef const struct _IConfigWidgetVtbl IConfigWidgetVtbl;
typedef struct _ConfigWidget ConfigWidget;
typedef char *Navigate(Dialog *pDialog, int offs);
typedef struct _ConfigGeneral ConfigGeneral;
typedef struct _ConfigCommon ConfigCommon;
typedef struct _ConfigWasapi ConfigWasapi;
typedef struct _ConfigYasapi ConfigYasapi;
typedef struct _ConfigMonitorPlayer ConfigMonitorPlayer;
typedef struct _ConfigMonitorEngine ConfigMonitorEngine;
typedef struct _ConfigMonitorState ConfigMonitorState;
typedef struct _ConfigMonitor ConfigMonitor;
typedef struct _ConfigFormatMatrix ConfigFormatMatrix;
typedef const struct _ClassCell ClassCell;

typedef const struct _IDialogVtbl IDialogVtbl;
typedef struct _Config Config;
typedef enum _EAboutLimits EAboutLimits;
typedef struct _About About;
typedef enum _EDialog EDialog;

typedef enum _EGtk EGtk;
typedef enum _EGtkLimits EGtkLimits;
typedef const struct _IFsmState IFsmState;
typedef struct _Fsm Fsm;
typedef struct _GtkServer GtkServer;
typedef struct _GtkClient GtkClient;
typedef const struct _IGtkClientState IGtkClientState;
typedef struct _GtkDialogs GtkDialogs;

///////////////////////////////////////////////////////////////////////////////
enum _EFormatMatrixLimits {
  eFrequenciesNum=8,
  eBitDepthsNum=3,
};

struct _FormatMatrix {
  int bSupported[eFrequenciesNum][eBitDepthsNum];
};

int FormatMatrixCreate(FormatMatrix *pMatrix, IMMDevice *pDevice);
void FormatMatrixDestroy(FormatMatrix *pMatrix);

///////////////////////////////////////////////////////////////////////////////
struct _OptionsGtk {
  wchar_t path[MAX_PATH];
};

extern const wchar_t gcaKeyGtk[];
extern const OptionsGtk gcOptionsGtk;

////////////////
struct _OptionsAbout {
  int nPosX;
  int nPosY;
  int nPage;
};

extern const OptionsAbout gcOptionsAbout;

struct _OptionsConfig {
  int nPosX;
  int nPosY;
  int nPage;

  struct {
    int nPage;
  } common;

  struct {
    int nPage;
  } monitor;
};

extern const OptionsConfig gcOptionsConfig;

////////////////
enum _EMono2Stereo {
  eMono2StereoOff,
  eMono2StereoPlain,
  eMono2StereoHalf,
  eMono2StereoSqrt2Inverse,
};



enum _EVolume {
  eVolumeOff,
  eVolumeSoftware,
  eVolumeWasapi,
};

enum _EDither {
  eDitherTruncate,
  eDitherRound,
  eDitherSimple,
};

enum _EConfigTheme {
  eConfigThemeDefault,
  eConfigThemeRed,
  eConfigThemeGreen,
  eConfigThemeBlue,
  eConfigThemeMax,
};

enum _EThreadPriority {
  eThreadPriorityOff,
  eThreadPriorityBackground,
  eThreadPriorityAboveNormal,
  eThreadPriorityBelowNormal,
  eThreadPriorityHighest,
  eThreadPriorityIdle,
  eThreadPriorityLowest,
  eThreadPriorityNormal,
  eThreadPriorityTimeCritical,
};

struct _OptionsCommon {
  wchar_t id[128];
  int bLatency;
  EMono2Stereo eMono2Stereo;
  EVolume eVolume;
  EDither eDither;
  int bSurround;
  int bDisconnect;
  int bGapless;
  EConfigTheme eThemeEngineTail;
  EConfigTheme eThemePadding;
  EConfigTheme eThemeStart;
  EConfigTheme eThemeWritten;

  struct {
    int bEnabled;
    double xPeriod;
  } present;

  struct {
    EThreadPriority ePriority;
  } winamp;

  struct {
    EThreadPriority ePriority;
  } drain;
};

extern const OptionsCommon gcOptionsCommon;

////////////////
enum _EBitdepth {
  eBitdepthDisabled=0,
  eBitdepth16=16,
  eBitdepth24=24,
  eBitdepth32=32,
};

enum _EPeriod {
  ePeriodDefault,
  ePeriodMinimum,
};

enum _EMode {
  eModeShared,
  eModeExclusive,
};

enum _EStrategy {
  eStrategyPush,
  eStrategyPull,
};


struct _OptionsDevice {
  EMode eMode;
  int bAutoConvertPCM;
  int bSRCDefaultQuality;
  EStrategy eStrategy;
  int bTimerPeriodic;
  int bAlignmentDance;
  int bProAudio;

  struct {
    EThreadPriority ePriority;
  } render;

  EBitdepth eBitdepth;
  EPeriod ePeriod;
  double xMaxLatencyAssumed;
  double xPlaybackBegin;
};

extern const OptionsDevice gcOptionsDevice;

////////////////
struct _OptionsMemberAbout {
  OptionsAbout current;
  OptionsAbout original;
};

struct _OptionsMemberConfig {
  OptionsConfig current;
  OptionsConfig original;
};

struct _OptionsMemberGtk {
  OptionsGtk current;
  OptionsGtk original;
};

////////////////
struct _OptionsMemberCommon {
  OptionsCommon config;
  OptionsCommon current;
  OptionsCommon original;
};

////////////////
struct _OptionsMemberDevice {
  const wchar_t *pId;
  UINT uRef;
  OptionsDevice config;
  OptionsDevice current;
  OptionsDevice original;
  FormatMatrix matrix;
};

OptionsMemberDevice *OptionsSearch(Options *pOptions, const wchar_t *id);
void OptionsMemberDeviceRef(OptionsMemberDevice *pDevice);
void OptionsMemberDeviceUnref(OptionsMemberDevice *pDevice,
    Options *pOptions);

////////////////
struct _Options {
  wchar_t *ini;
  OptionsMemberCommon common;
  OptionsMemberDevice *pDevices;
  OptionsMemberDevice *pDevicesMax;
  OptionsMemberDevice *pDevicesAlloc;
};

int OptionsCreate(Options *pOptions, Player *pPlayer, UINT cDevices);
void OptionsDestroy(Options *pOptions);

void OptionsRollback(Options *pOptions);
void OptionsSave(Options *pOptions);

///////////////////////////////////////////////////////////////////////////////
struct _ICacheDeviceVtbl {
  const wchar_t *(*getid)(CacheDevice *pDevice);
};

struct _CacheDevice {
  ICacheDeviceVtbl *lpVtbl;
  IMMDevice *pDevice;
  wchar_t *id;
  IPropertyStore *pStore;
  PROPVARIANT vFriendlyName;
  char *pFriendlyNameOriginal;
  char *pFriendlyName;
  char *pFriendlyNameMarkup;
  OptionsMemberDevice *pMember;
};

int CacheDeviceCreate(CacheDevice *pDevice, Cache *pCache, int nDevice);
void CacheDeviceDestroy(CacheDevice *pDevice, Options *pOptions);

///////////////////////////////////////////////////////////////////////////////
struct _Cache {
  Player *pPlayer;
  IMMDeviceEnumerator *pEnumerator;
  IMMDeviceCollection *pCollection;
  Options options;

  struct {
    struct {
      CacheDevice *pCacheDevice;
    } active;

    struct {
      CacheDevice *pCacheDevice;
    } config;
  } selected;

  UINT cCacheDevices;
  CacheDevice *pCacheDevices;
};

int CacheCreate(Cache *pCache, Player *pPlayer);
void CacheDestroy(Cache *pCache);

int CacheSelectDevice(Cache *pCache, CacheDevice *pDevice);
void CacheRestore(Cache *pCache);
void CacheSync(Cache *pCache);
void CacheDefault(Cache *pCache);

///////////////////////////////////////////////////////////////////////////////
struct _IPropertyVtbl {
  void (*read)(const Property *pProperty, const wchar_t *path,
      const wchar_t *lpAppName, char *data, const char *ref);
  void (*write)(const Property *pProperty, const wchar_t *path,
      const wchar_t *lpAppName, const char *data, const char *ref);
};

struct _Property {
  IPropertyVtbl *lpVtbl;
  const wchar_t *lpKeyName;
  int offs;
  DWORD dwSize;
};

extern const Property gcaPropertiesAbout[];
extern const Property gcaPropertiesConfig[];
extern const Property gcaPropertiesGtk[];
extern const Property gcaPropertiesCommon[];
extern const Property gcaPropertiesDevice[];

void PropertiesRead(const Property *pProperty, const wchar_t *path,
    const wchar_t *lpAppName, void *data, const void *ref);
void PropertiesWrite(const Property *pProperty, const wchar_t *path,
    const wchar_t *lpAppName, const void *data, const void *ref);

///////////////////////////////////////////////////////////////////////////////
struct _Frequency {
  const char *label;
  int freq;
};

extern Frequency gaFrequencies[eFrequenciesNum];

struct _BitDepth {
  const char *label;
  int bits;
};

extern BitDepth gaBitDepths[eBitDepthsNum];

struct _IConfigWidgetVtbl {
  void (*set)(ConfigWidget *pWidget);
  void (*get)(ConfigWidget *pWidget);
};

struct _ConfigWidget {
  IConfigWidgetVtbl *lpVtbl;
  Dialog *pDialog;
  ClassCell *pCell;
  GtkWidget *widget;
  GtkCssProvider *provider;
};

struct _ConfigGeneral {
  ConfigWidget device;
  ConfigWidget mode;
  ConfigWidget strategy;
};

struct _ConfigCommon {
  ConfigWidget latency;
  ConfigWidget mono2stereo;
  ConfigWidget volume;
  ConfigWidget dither;
  ConfigWidget surround;
  ConfigWidget disconnect;
  ConfigWidget gapless;

  struct {
    ConfigWidget engine_tail;
    ConfigWidget padding;
    ConfigWidget start;
    ConfigWidget written;
  } theme;

  struct {
    ConfigWidget enabled;
    ConfigWidget period;
  } present;

  struct {
    ConfigWidget priority;
  } winamp;

  struct {
    ConfigWidget priority;
  } drain;
};

struct _ConfigWasapi {
  ConfigWidget device;
  ConfigWidget mode;
  ConfigWidget auto_convert_pcm;
  ConfigWidget src_default_quality;
  ConfigWidget strategy;
  ConfigWidget timer_periodic;
  ConfigWidget pro_audio;

  struct {
    ConfigWidget priority;
  } render;

  ConfigWidget alignment_dance;
  ConfigWidget period;
};

struct _ConfigYasapi {
  ConfigWidget device;
  ConfigWidget bitdepth;
  ConfigWidget playback_begin;
  ConfigWidget period_playback_begin;
  ConfigWidget max_latency_assumed;
};

struct _ConfigMonitorPlayer {
  ConfigWidget num_tracks;
  ConfigWidget num_gapless;
  ConfigWidget ms_dropout;
  ConfigWidget ms_latency;
  ConfigWidget num_gapless_broken;
};

struct _ConfigMonitorState {
  ConfigWidget client;
  ConfigWidget server;
};

struct _ConfigMonitorEngine {
  ConfigWidget volume;
  ConfigWidget num_tracks;
  ConfigWidget num_frames_device;
  ConfigWidget num_frames_ring;
  ConfigWidget num_frames_latency;
  ConfigWidget num_frames_dropout;
  ConfigWidget tail_max;
  ConfigWidget tail_min;
  ConfigWidget tail_cur;
};

struct _ConfigMonitor {
  ConfigWidget device;

  ConfigMonitorPlayer player;
  ConfigMonitorState state;
  ConfigMonitorEngine engine;
  ConfigWidget padding;
  ConfigWidget written;
  ConfigWidget start;
};

struct _ConfigFormatMatrix {
  ConfigWidget device;
  ConfigWidget enabled[eFrequenciesNum][eBitDepthsNum];
};

///////////////////////////////////////////////////////////////////////////////
struct _IDialogVtbl {
  const wchar_t *pIdW;
  const char *pTitle;
  int nWidthPixels;
  void (*WindowCreate)(Dialog *pDialog);
  void (*set)(Dialog *pDialog);
  void (*get)(Dialog *pDialog);
  int (*PropertiesRead)(Dialog *pDialog, Player *pPlayer);
  int (*PropertiesWrite)(Dialog *pDialog, Player *pPlayer);
  void (*Finalize)(Dialog *pDialog);
};

struct _Config {
  union {
    struct {
      ConfigGeneral general;
      ConfigCommon common;
      ConfigWasapi wasapi;
      ConfigYasapi yasapi;
      ConfigMonitor monitor;
      ConfigFormatMatrix explore;
    };

    ConfigWidget widgets[1];
  };

  struct {
    GtkWidget *root;
    GtkWidget *common;
    GtkWidget *monitor;
  } notebook;
  
  OptionsMemberConfig options;
};

enum _EAboutLimits {
  eAboutLimitsURLSize = 128,
  eAboutLimitsLinksSize = 30,
  eAboutLimitsPagesNum = 3,
  //eAboutLimitsWidthPixels=455,
  eAboutLimitsWidthPixels=492,
  //eAboutLimitsWidthPixels=518,
  eAboutLimitsHeightPixels=296,
  eAboutLimitsWidthChars=77,
};

struct _About {
#if defined (YASAPI_ABOUT_TEXT_VIEW) // [
  GdkDisplay *display;
  GdkCursor *arrow;
  GdkCursor *hand;
  AboutPage *pPageActive;
  AboutPage pages[eAboutLimitsPagesNum];
#else // ] [
  GtkWidget *notebook;
  OptionsMemberAbout options;
#endif // ]
};

extern IDialogVtbl gIAboutVtbl;
extern IDialogVtbl gIConfigVtbl;

enum _EDialog {
  eAboutWidthPixels=492,
  eConfigWidthPixels=680,
};

struct _Dialog {
  IDialogVtbl *lpVtbl;
  HWND hWndParent;
  GtkClient *pGtk;
  GtkWidget *window;

  
  int bMoved;

  union {
    About about;
    Config config;
  };
};

OptionsCommon *DialogOptionsCommonConfig(Dialog *pDialog);
OptionsDevice *DialogOptionsDeviceConfig(Dialog *pDialog);

void GtkServerDialogWrite(Dialog *pDialog);
void GtkServerDialogShowSync (Dialog *pDialog);
gboolean GtkServerDialogShowAsync (gpointer user_data);
gboolean GtkServerPresentAsync (gpointer user_data);
void gtk_dialog_options_save (GtkWidget *widget, gpointer data);
gboolean window_delete_event (GtkWidget *window, GdkEvent *event,
    gpointer user_data);

////////////////
enum _EGtkLimits {
  eGtkNumDialogs=2,
};


///////////////////////////////////////////////////////////////////////////////
struct _Fsm {
  IFsmState *pStateMin;
  IFsmState *pState;
  GtkClient *pGtkClient;
  ConfigWidget *pWidget;
  const char *pId;
};

void FsmCreate(void *pe, const void *ps , GtkClient *pGtkClient, ConfigWidget *pWidget);
void FsmDestroy(void *pe);

int FsmState(void *pe);
const char *FsmStateIdEx(void *pe, int nState);
const char *FsmStateId(void *pe);
int FsmBacktrack(void *pe, int nState, const char *file, int line, const char *func);
int FsmUp(void *pe, int nState, const char *file, int line, const char *func);
void FsmDown(void *pe, int nState );

///////////////////////////////////////////////////////////////////////////////
struct _GtkClient {
  Fsm fsm;
  Player *pPlayer;
  GtkServer *pServer;
  Dialog *pDialog;
  HWND hWndParent;
  WNDPROC lpPrevWndFunc;

  union {
    struct {
      Dialog about;
      Dialog config;
    };

    Dialog aDialog[eGtkNumDialogs];
  };

  HMODULE hLibGlib;

  struct {
    HANDLE hEvent;
    HANDLE hThread;
    DWORD dwThreadId;
  } thread;
};

int GtkClientCreate(GtkClient *pClient, Player *pPlayer);
void GtkClientDestroy(GtkClient *pClient);

int GtkClientShow(GtkClient *pClient, Dialog *pDialog, HWND hWndParent);

void GtkClientUpdateWidget(GtkClient *pClient, ConfigWidget *pWidget);
void GtkClientUpdateDialog(GtkClient *pClient, Dialog *pDialog);

///////////////////////////////////////////////////////////////////////////////
char *yasapi_wcs2strdup(wchar_t *wcs, UINT uCodePage);
void yasapi_message(const char *file, int line, const char *format,
    DWORD dwError, ...);
const wchar_t *yasapi_basenamew(const wchar_t *path);
const char *yasapi_basenamea(const char *path);
int yasapi_wcsstrcmp(const wchar_t *wcs, const char *str);

#ifdef __cplusplus // [
}
#endif // ]
#endif // __YASAPI_PARAM_H__ ]
