/*
 * yasapi_present_client_state.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
static int PresentClientStateEventEnter(PresentClient *pClient)
{
  // This is a MANUAL RESET event!
  if (!(pClient->hEvent=CreateEventW(NULL,TRUE,FALSE,NULL))) {
    DMESSAGE("creating consumer event");
    goto exit;
  }

  ++pClient->fsm.pState;

  return 0;
exit:
  return -1;
}

static void PresentClientStateEventLeave(PresentClient *pClient)
{
  --pClient->fsm.pState;
  CloseHandle(pClient->hEvent);
}

///////////////////////////////////////////////////////////////////////////////
static int PresentClientStateThreadEnter(PresentClient *pClient)
{
  pClient->hThread=CreateThread(
    NULL,       // _In_opt_  LPSECURITY_ATTRIBUTES  lpThreadAttributes,
    0ul,        // _In_      SIZE_T                 dwStackSize,
    PresentServerThreadProc,
                // _In_      LPTHREAD_START_ROUTINE lpStartAddress,
    pClient,  // _In_opt_  LPVOID                 lpParameter,
    0ul,        // _In_      DWORD                  dwCreationFlags,
                // _Out_opt_ LPDWORD                lpThreadId
    &pClient->dwThreadId
  );

  if (INVALID_HANDLE_VALUE==pClient->hThread) {
    DMESSAGE("starting present server");
    goto exit;
  }

  ++pClient->fsm.pState;

  return 0;
exit:
  return -1;
}

static void PresentClientStateThreadLeave(PresentClient *pClient)
{
  --pClient->fsm.pState;
  WaitForSingleObject(pClient->hThread,INFINITE);
  CloseHandle(pClient->hThread);
  pClient->hThread=INVALID_HANDLE_VALUE;
}

///////////////////////////////////////////////////////////////////////////////
static int PresentClientStateRunningEnter(PresentClient *pClient)
{
  IFsmState *pState=pClient->fsm.pStateMin+ePresentClientStateRunning;

  while (pClient->fsm.pState<pState) {
    ResetEvent(pClient->hEvent);
    WaitForSingleObject(pClient->hEvent,INFINITE);
  }

  DWRITELN("!!! THREAD RUNNING !!!!!!!!!!!!!!!!!!!!!!");

  return 0;
}

static void PresentClientStateRunningLeave(PresentClient *pClient)
{
  IFsmState *pState=pClient->fsm.pStateMin+ePresentClientStateRunning-1;
  PresentServer *pServer=pClient->pServer;

  DASSERT(pServer,exit);
  pServer->eStateRequest=ePresentServerStateLoopEnd;
  SetEvent(pServer->hEvent);

  while (pState<pClient->fsm.pState) {
    ResetEvent(pClient->hEvent);
    WaitForSingleObject(pClient->hEvent,INFINITE);
  }

DLABEL(exit);
  return;
}

///////////////////////////////////////////////////////////////////////////////
IPresentClientState gcaIPresentClientState[ePresentClientStateCount]={
#define YASAPI_PRESENT_CLIENT_STATE(x) { \
  .pId=#x, \
  .Enter=PresentClientState ## x ## Enter, \
  .Leave=PresentClientState ## x ## Leave, \
},
#include <yasapi_present_client_state.h>
#undef YASAPI_PRESENT_CLIENT_STATE
};
