/*
 * yasapi_totop_server.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>
#include <winamp/wa_ipc.h>

///////////////////////////////////////////////////////////////////////////////
static int PresentServerCreate(PresentServer *pServer, PresentClient *pClient);
static void PresentServerDestroy(PresentServer *pServer);
static void PresentServerHandshake(PresentServer *pServer);
//static void PresentServerError(PresentServer *pServer);

///////////////////////////////////////////////////////////////////////////////
DWORD WINAPI PresentServerThreadProc(LPVOID lpParameter)
{
  enum {
    eEvent,
    eTimer,
    eCount,
  };

  PresentServer server;
  Player *pPlayer;
  HANDLE aHandles[eCount];
  DWORD dwWait;

  if (PresentServerCreate(&server,lpParameter)<0)
    goto ecreate;

  if (YASAPI_FSM_UP(&server,ePresentServerStateTimerStart)<0)
    goto eopen;

  pPlayer=server.pClient->pPlayer;
  aHandles[eEvent]=server.hEvent;
  aHandles[eTimer]=server.hTimer;
  ++server.pClient->fsm.pState;
  SetEvent(server.pClient->hEvent);
  SwitchToThread();

  for (;;) {
    ResetEvent(server.hEvent);

    dwWait=WaitForMultipleObjects(
      eCount,       // _In_       DWORD  nCount,
      aHandles,     // _In_ const HANDLE *lpHandles,
      FALSE,        // _In_       BOOL   bWaitAll,
      INFINITE      // _In_       DWORD  dwMilliseconds
    );

    switch (dwWait) {
    case WAIT_OBJECT_0+eEvent:
      switch (server.eStateRequest) {
      case ePresentServerStateLoopEnd:
        goto end;
      case ePresentServerStateTimerStart:
        if (FsmState(&server)<server.eStateRequest)
          YASAPI_FSM_UP(&server,server.eStateRequest);
        else
          YASAPI_FSM_BACKTRACK(&server,server.eStateRequest);

        break;
      case ePresentServerStateBase:
        YASAPI_FSM_DOWN(&server,server.eStateRequest);
        break;
      default:
        break;
      }

      PresentServerHandshake(&server);
      continue;
    case WAIT_OBJECT_0+eTimer:
        if (SendMessage(pPlayer->pOut->hMainWindow,WM_WA_IPC,0,IPC_ISVISRUNNING))
          g_idle_add (GtkServerPresentAsync, server.pClient->pPlayer->gtk.pServer);

      continue;
    case WAIT_ABANDONED_0+eEvent:
      DWRITELN("WAIT_ABANDONED eEvent");
      continue;
    case WAIT_ABANDONED_0+eTimer:
      DWRITELN("WAIT_ABANDONED eTimer");
      continue;
    case WAIT_TIMEOUT:
      DWRITELN("WAIT_TIMEOUT");
      continue;
    case WAIT_FAILED:
      DVWRITELN("WAIT_FAILED: %lu",GetLastError());
      continue;
    default:
      DVWRITELN("unexpectedly received %lu",dwWait);
      continue;
    }
  }
end:
  --server.pClient->fsm.pState;
  PresentServerHandshake(&server);
eopen:
  PresentServerDestroy(&server);
ecreate:
  ExitThread(0ul);
  return 0ul;
}

///////////////////////////////////////////////////////////////////////////////
static int PresentServerCreate(PresentServer *pServer, PresentClient *pClient)
{
  memset(pServer,0,sizeof *pServer);
  FsmCreate(pServer,gcaIPresentServerState ,NULL,NULL);
  pServer->pClient=pClient;
  pClient->pServer=pServer;

  return 0;
}

static void PresentServerDestroy(PresentServer *pServer)
{
  pServer->pClient->pServer=NULL;
  FsmDestroy(pServer);
}

///////////////////////////////////////////////////////////////////////////////
static void PresentServerHandshake(PresentServer *pServer)
{
  pServer->eStateRequest=ePresentServerStateStart;
  SetEvent(pServer->pClient->hEvent);
}
